# vim: ts=4:sts=4:sw=4
#
# @author: <sylvain.herledan@oceandatalab.com>
# @date: 2016-10-11
#
# This file is part of SEAScope, a 3D visualisation and analysis application
# for satellite, in-situ and numerical model data.
#
# Copyright (C) 2014-2023 OceanDataLab
#
# SEAScope is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SEAScope. If not, see <https://www.gnu.org/licenses/>.

"""
This module handles the serialization and deserialization of IDF descriptor
objects
"""

import logging
import flatbuffers
import SEAScope.API.IDFDescriptor
import SEAScope.types.shape

logger = logging.getLogger(__name__)


def serialize(builder, obj):
    """
    Serialize an IDF descriptor using FlatBuffers.

    Parameters
    ----------
    builder : flatbuffers.builder.Builder
        The FlatBuffers builder instance which serializes data. If this
        parameter is None, then a new builder will be created
    obj : dict
        Dictionary which contains information about the IDF descriptor to
        serialize

        It must have 6 keys:

        - ``uri`` : a :obj:`str` which is the address where SEAScope will find
          the granule

        - ``resolution`` : a :obj:`float` which is the resolution, in meters
          per cell of data, of the granule

        - ``subsampling_factor`` : an :obj:`int` which tells SEAScope how many
          times the data contained in the granule have been subsampled (by 2)
          compared to the full resolution data

        - ``xArity`` : an :obj:`int` which tells SEAScope how many columns it
          must allocate when reconstructing the data matrix

        - ``yArity`` : an :obj:`int` which tells SEAScope how many rows it must
          allocate when reconstructing the data matrix

        - ``shape`` : a :obj:`dict` that describes the shape of the granule.
          The value must satisfy the requirements of the
          :func:`SEAScope.types.shape.serialize` method

    Returns
    -------
    tuple(flatbuffers.builder.Builder, int)
        A tuple which contains two elements:

        - the :obj:`flatbuffers.builder.Builder` instance which has been used
          to serialize data
        - an :obj:`int` which is the address/offset of the serialized object
          in the builder buffer
    """
    if builder is None:
        builder = flatbuffers.Builder(0)

    uri = builder.CreateString(obj['uri'])
    resolution = obj['resolution']
    subsampling_factor = int(obj['subsampling_factor'])
    x_arity = int(obj['xArity'])
    y_arity = int(obj['yArity'])
    _, shape = SEAScope.types.shape.serialize(builder, obj['shape'])

    cls = SEAScope.API.IDFDescriptor
    cls.IDFDescriptorStart(builder)
    cls.IDFDescriptorAddUri(builder, uri)
    cls.IDFDescriptorAddResolution(builder, resolution)
    cls.IDFDescriptorAddXArity(builder, x_arity)
    cls.IDFDescriptorAddYArity(builder, y_arity)
    cls.IDFDescriptorAddShape(builder, shape)
    cls.IDFDescriptorAddSubsamplingFactor(builder, subsampling_factor)
    idf_descriptor = cls.IDFDescriptorEnd(builder)

    return builder, idf_descriptor


def deserialize(o):
    """
    Rebuild an IDF descriptor from a FlatBuffers buffer.

    Parameters
    ----------
    buf : bytearray
        The buffer which contain the IDF descriptor object serialized with
        FlatBuffers

    Returns
    -------
    dict
        The deserialized IDF descriptor object as a dictionary.
    """
    result = {}
    result['uri'] = o.Uri()
    result['resolution'] = o.Resolution()
    result['xArity'] = o.XArity()
    result['yArity'] = o.YArity()
    result['shape'] = SEAScope.types.shape.deserialize(o.Shape())
    result['subsampling_factor'] = o.SubsamplingFactor()
    return result



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: API

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class ListSourcesResponse(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsListSourcesResponse(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = ListSourcesResponse()
        x.Init(buf, n + offset)
        return x

    # ListSourcesResponse
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ListSourcesResponse
    def Sources(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from SEAScope.API.Source import Source
            obj = Source()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # ListSourcesResponse
    def SourcesLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # ListSourcesResponse
    def SourcesIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

def ListSourcesResponseStart(builder): builder.StartObject(1)
def ListSourcesResponseAddSources(builder, sources): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(sources), 0)
def ListSourcesResponseStartSourcesVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def ListSourcesResponseEnd(builder): return builder.EndObject()



# vim: ts=4:sts=4:sw=4
#
# @author: <sylvain.herledan@oceandatalab.com>
# @date: 2016-09-07
#
# This file is part of SEAScope, a 3D visualisation and analysis application
# for satellite, in-situ and numerical model data.
#
# Copyright (C) 2014-2023 OceanDataLab
#
# SEAScope is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SEAScope. If not, see <https://www.gnu.org/licenses/>.

"""
This module handles the serialization and deserialization of timespan objects
"""

import logging
import flatbuffers
import SEAScope.API.Timespan

logger = logging.getLogger(__name__)


def serialize(builder, span_obj):
    """
    Serialize a timespan using FlatBuffers.

    Parameters
    ----------
    builder : flatbuffers.builder.Builder
        The FlatBuffers builder instance which serializes data. If this
        parameter is None, then a new builder will be created
    span_obj : dict
        Dictionary which contains information about the timespan to serialize
        It must have three keys:

        - ``label`` : a :obj:`str` displayed by SEAScope to designate this
          timespan

        - ``pastOffset`` : an :obj:`int` that tells SEAScope when the timespan
          starts, as a positive number of seconds relative to SEAScope current
          datetime, i.e. ``start = current - pastOffset``

        - ``futureOffset`` : an :obj:`int` that tells SEAScope when the
          timespan ends, as a positive number of seconds relative to SEAScope
          current datetime, i.e. ``end = current + futureOffset``

    Returns
    -------
    tuple(flatbuffers.builder.Builder, int)
        A tuple which contains two elements:

        - the :obj:`flatbuffers.builder.Builder` instance which has been used
          to serialize data
        - an :obj:`int` which is the address/offset of the serialized object
          in the builder buffer
    """
    if builder is None:
        builder = flatbuffers.Builder(0)

    _label = builder.CreateString(span_obj['label'])
    _past = span_obj['pastOffset']
    _future = span_obj['futureOffset']
    cls = SEAScope.API.Timespan
    cls.TimespanStart(builder)
    cls.TimespanAddLabel(builder, _label)
    cls.TimespanAddPastOffset(builder, _past)
    cls.TimespanAddFutureOffset(builder, _future)
    timespan = cls.TimespanEnd(builder)

    return builder, timespan


def deserialize(o):
    """
    Rebuild a timespan from a FlatBuffers buffer.

    Parameters
    ----------
    buf : bytearray
        The buffer which contains the timespan object serialized with
        FlatBuffers

    Returns
    -------
    dict
        The deserialized timespan object as a dictionary.
    """
    result = {}
    result['label'] = o.Label()
    result['pastOffset'] = o.PastOffset()
    result['futureOffset'] = o.FutureOffset()
    return result



# vim: ts=4:sts=4:sw=4
#
# @author: <sylvain.herledan@oceandatalab.com>
# @date: 2016-09-09
#
# This file is part of SEAScope, a 3D visualisation and analysis application
# for satellite, in-situ and numerical model data.
#
# Copyright (C) 2014-2023 OceanDataLab
#
# SEAScope is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SEAScope. If not, see <https://www.gnu.org/licenses/>.

"""
This module handles the serialization and deserialization of color values
"""

import logging
import flatbuffers
import SEAScope.API.Color

logger = logging.getLogger(__name__)


def serialize(builder, color_obj):
    """
    Serialize a rendering configuration using FlatBuffers.

    Parameters
    ----------
    builder : flatbuffers.builder.Builder
        The FlatBuffers builder instance which serializes data. If this
        parameter is None, then a new builder will be created
    color_obj : :obj:`list` of 3 :obj:`int`
        The color described by its red, green and blue channels

    Returns
    -------
    tuple(flatbuffers.builder.Builder, int)
        A tuple which contains two elements:

        - the :obj:`flatbuffers.builder.Builder` instance which has been used
          to serialize data
        - an :obj:`int` which is the address/offset of the serialized object
          in the builder buffer
    """
    if builder is None:
        builder = flatbuffers.Builder(0)

    cls = SEAScope.API.Color
    color = cls.CreateColor(builder,
                            int(color_obj[0]),
                            int(color_obj[1]),
                            int(color_obj[2]))

    return builder, color


def deserialize(o):
    """
    Rebuild a color from a FlatBuffers buffer.

    Parameters
    ----------
    buf : bytearray
        The buffer which contains the color object serialized with FlatBuffers

    Returns
    -------
    dict
        The deserialized color object as a :obj:`list` of 3 :obj:`int` (R, G
        and B)
    """
    result = []
    result.append(o.R())
    result.append(o.G())
    result.append(o.B())
    return result



# vim: ts=4:sts=4:sw=4
#
# @author: <sylvain.herledan@oceandatalab.com>
# @date: 2016-09-07
#
# This file is part of SEAScope, a 3D visualisation and analysis application
# for satellite, in-situ and numerical model data.
#
# Copyright (C) 2014-2023 OceanDataLab
#
# SEAScope is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SEAScope. If not, see <https://www.gnu.org/licenses/>.

"""
This module handles the serialization and deserialization of collection objects
"""

import logging
import flatbuffers
import SEAScope.API.Collection
import SEAScope.types.renderable_id
import SEAScope.types.tag
import SEAScope.types.variable

logger = logging.getLogger(__name__)


def serialize(builder, collection_obj):
    """
    Serialize a collection using FlatBuffers.

    Parameters
    ----------
    builder : flatbuffers.builder.Builder
        The FlatBuffers builder instance which serializes data. If this
        parameter is None, then a new builder will be created
    collection_obj : dict
        Dictionary which contains information about the collection to serialize
        It must have ten keys:

        - ``mustBeCurrent`` : a :obj:`bool` which tells SEAScope how its
          temporal filter must be applied for this collection.

          When this setting is set to true, the viewer will ignore granules
          whose time range does not contain the current datetime.
          It is the recommended behavior for collections whose granules cover
          the whole globe.

          When it is set to false, the viewer displayed the granules of the
          collection as long as their time range intersects the current time
          window (cf. timespans in the list of settings for the application
          configuration file).
          Use this behavior for collections of granules with a smaller spatial
          footprint or a sparse temporal coverage.

        - ``xSeamless`` : a :obj:`bool` which tells SEAScope that the granules
          of the collection have a 360° longitudinal cover

        - ``ySeamless`` : a :obj:`bool` which tells SEAScope that the granules
          of the collection have a 180° latitudinal cover

        - ``NEWSAligned`` : a :obj:`bool` which tells SEAScope that the axes of
          the data matrix are aligned witth the South-North and West-East axes

        - ``label`` : a :obj:`str` that SEAScope uses to designate the
          collection

        - ``tags`` : a :obj:`dict` which contains tags that describe the
          collection. SEAScope considers a (key, value) pair as a tag.

        - ``variables`` : a :obj:`list` of :obj:`dict` containing information
          about the variables defined for the collection.
          Each :obj:`dict` objects must have all the keys required by
          :func:`SEAScope.types.variable.serialize`

        - ``defaultVariable`` : a :obj:`int` which is the index of the default
          item in the ``variables`` list

        - ``id`` : a :obj:`dict` whose content can be used by SEAScope as a
          unique identifier for  the collection.
          It must have all the keys required by
          :func:`SEAScope.types.renderable_id.serialize`

    Returns
    -------
    tuple(flatbuffers.builder.Builder, int)
        A tuple which contains two elements:

        - the :obj:`flatbuffers.builder.Builder` instance which has been used
          to serialize data
        - an :obj:`int` which is the address/offset of the serialized object
          in the builder buffer
    """
    if builder is None:
        builder = flatbuffers.Builder(0)

    cls = SEAScope.API.Collection

    _must_be_current = collection_obj['mustBeCurrent']
    _x_seamless = collection_obj['xSeamless']
    _y_seamless = collection_obj['ySeamless']
    _news_aligned = collection_obj['NEWSAligned']
    _label = builder.CreateString(collection_obj['label'])

    tags = []
    tags_len = len(collection_obj['tags'].keys())
    for k, v in collection_obj['tags'].items():
        _, tag = SEAScope.types.tag.serialize(builder, k, v)
        tags.append(tag)
    cls.CollectionStartTagsVector(builder, tags_len)
    for t in tags:
        builder.PrependUOffsetTRelative(t)
    _tags = builder.EndVector(tags_len)

    variables = []
    vars_len = len(collection_obj['variables'])
    for var_obj in collection_obj['variables']:
        _, variable = SEAScope.types.variable.serialize(builder, var_obj)
        variables.append(variable)
    cls.CollectionStartVariablesVector(builder, vars_len)
    for v in variables:
        builder.PrependUOffsetTRelative(v)
    _variables = builder.EndVector(vars_len)

    _default_variable = int(collection_obj['defaultVariable'])
    _, _rid = SEAScope.types.renderable_id.serialize(builder,
                                                     collection_obj['id'])

    cls.CollectionStart(builder)
    cls.CollectionAddId(builder, _rid)
    cls.CollectionAddMustBeCurrent(builder, _must_be_current)
    cls.CollectionAddXSeamless(builder, _x_seamless)
    cls.CollectionAddYSeamless(builder, _y_seamless)
    cls.CollectionAddNewsAligned(builder, _news_aligned)
    cls.CollectionAddLabel(builder, _label)
    cls.CollectionAddTags(builder, _tags)
    cls.CollectionAddVariables(builder, _variables)
    cls.CollectionAddDefaultVariable(builder, _default_variable)
    collection = cls.CollectionEnd(builder)

    return builder, collection


def deserialize(buf):
    """
    Rebuild a collection from a FlatBuffers buffer.

    Parameters
    ----------
    buf : bytearray
        The buffer which contains the collection object serialized with
        FlatBuffers

    Returns
    -------
    dict
        The deserialized collection object as a dictionary. Note that the
        "defaultVariable" key is lost in the process.
    """
    col_obj = {}
    col_obj['id'] = SEAScope.types.renderable_id.deserialize(buf.Id())
    col_obj['label'] = buf.Label()
    col_obj['mustBeCurrent'] = 0 < buf.MustBeCurrent()
    col_obj['xSeamless'] = 0 < buf.XSeamless()
    col_obj['ySeamless'] = 0 < buf.YSeamless()
    col_obj['NEWSAligned'] = 0 < buf.NewsAligned()
    col_obj['tags'] = {}
    tags_count = buf.TagsLength()
    for i in range(0, tags_count):
        name, value = SEAScope.types.tag.deserialize(buf.Tags(i))
        col_obj['tags'][name] = value

    col_obj['variables'] = []
    vars_count = buf.VariablesLength()
    for i in range(0, vars_count):
        v = SEAScope.types.variable.deserialize(buf.Variables(i))
        col_obj['variables'].append(v)

    return col_obj



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: API

class RenderingMethod(object):
    RASTER = 0
    ARROWS = 1
    BARBS = 2
    TRAJECTORIES = 3
    DOTSCLOUD = 4
    STREAMLINES = 5
    RAWRGB = 6




# automatically generated by the FlatBuffers compiler, do not modify

# namespace: API

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class GranuleData(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsGranuleData(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = GranuleData()
        x.Init(buf, n + offset)
        return x

    # GranuleData
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # GranuleData
    def Fields(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return ""

    # GranuleData
    def FieldsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # GranuleData
    def FieldsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # GranuleData
    def Buckets(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from SEAScope.API.DataBucket import DataBucket
            obj = DataBucket()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # GranuleData
    def BucketsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # GranuleData
    def BucketsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # GranuleData
    def FloatFields(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return ""

    # GranuleData
    def FloatFieldsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # GranuleData
    def FloatFieldsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

    # GranuleData
    def FloatBuckets(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from SEAScope.API.FloatDataBucket import FloatDataBucket
            obj = FloatDataBucket()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # GranuleData
    def FloatBucketsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # GranuleData
    def FloatBucketsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        return o == 0

def GranuleDataStart(builder): builder.StartObject(4)
def GranuleDataAddFields(builder, fields): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(fields), 0)
def GranuleDataStartFieldsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def GranuleDataAddBuckets(builder, buckets): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(buckets), 0)
def GranuleDataStartBucketsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def GranuleDataAddFloatFields(builder, floatFields): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(floatFields), 0)
def GranuleDataStartFloatFieldsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def GranuleDataAddFloatBuckets(builder, floatBuckets): builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(floatBuckets), 0)
def GranuleDataStartFloatBucketsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def GranuleDataEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: API

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Timestep(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsTimestep(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Timestep()
        x.Init(buf, n + offset)
        return x

    # Timestep
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Timestep
    def Label(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Timestep
    def Step(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

def TimestepStart(builder): builder.StartObject(2)
def TimestepAddLabel(builder, label): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(label), 0)
def TimestepAddStep(builder, step): builder.PrependUint64Slot(1, step, 0)
def TimestepEnd(builder): return builder.EndObject()



# vim: ts=4:sts=4:sw=4
#
# @author: <sylvain.herledan@oceandatalab.com>
# @date: 2016-09-09
#
# This file is part of SEAScope, a 3D visualisation and analysis application
# for satellite, in-situ and numerical model data.
#
# Copyright (C) 2014-2023 OceanDataLab
#
# SEAScope is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SEAScope. If not, see <https://www.gnu.org/licenses/>.

"""
This module handles the serialization and deserialization of granule metadata
objects
"""

import logging
import flatbuffers
import SEAScope.API.GranuleMetadata
import SEAScope.API.DataModel
import SEAScope.types.idf_descriptor

logger = logging.getLogger(__name__)
data_models = {'LAT_LON': SEAScope.API.DataModel.DataModel.LAT_LON,
               'Y_X': SEAScope.API.DataModel.DataModel.Y_X,
               'NJ_NI': SEAScope.API.DataModel.DataModel.NJ_NI,
               'ROW_CELL': SEAScope.API.DataModel.DataModel.ROW_CELL,
               'TIME_STATION': SEAScope.API.DataModel.DataModel.TIME_STATION,
               'TIME': SEAScope.API.DataModel.DataModel.TIME}
"""dict: Data models supported by SEAScope.
"""


def serialize(builder, obj):
    """
    Serialize a granule metadata using FlatBuffers.

    Parameters
    ----------
    builder : flatbuffers.builder.Builder
        The FlatBuffers builder instance which serializes data. If this
        parameter is None, then a new builder will be created
    obj : dict
        Dictionary which contains information about the shape to serialize
        It must have between 8 and 21 keys:

        - ``sourceId`` : an :obj:`int` which is the unique identifier for the
          source which provides the granule

        - ``collectionId`` : an :obj:`int` which is the unique identifier
          (within the data source) for the collection that owns the granule

        - ``granuleId`` : an :obj:`int` which is the unique identifier (within
          the data source) for the granule

        - ``dataId`` : a :obj:`str` which is the name of the granule. As for
          the ``granuleId`` it should remain unique within a data source

        - ``dataModel`` : a :obj:`str` that identifies the data model used to
          store the granule data. The value must be a key of the
          :const:`SEAScope.types.granule_metadata.data_models` dictionary

        - ``start`` : an :obj:`int` which is the first time when data is
          available in the granule, in milliseconds since
          1970-01-01T00:00:00.000Z

        - ``stop`` : an :obj:`int` which is the first time after ``start`` when
          data is not available anymore, in milliseconds since
          1970-01-01T00:00:00.000Z

        - ``uris`` : a :obj:`list` of :obj:`dict`. The elements of the list
          describe the files attached to the granule and must satisfy the
          requirements of the :func:`SEAScope.types.idf_descriptor.serialize`
          method

        - ``title`` : an optional :obj:`str` for the title/long name of the
          granule

        - ``institution`` : an optional :obj:`str` for the name of the
          institution which provided the granule

        - ``comment`` : an optional :obj:`str` for any comment about the
          granule

        - ``file_id`` : an optional :obj:`str` for an identifier or path of the
          original file which has been converted to IDF

        - ``product_version``: an optional :obj:`str` for the version of the
          product

        - ``lon_min`` : an optional :obj:`float` for the min longitude of the
          spatial coverage

        - ``lon_max`` : an optional :obj:`float` for the max longitude of the
          spatial coverage

        - ``lat_min`` : an optional :obj:`float` for the min latitude of the
          spatial coverage

        - ``lat_max`` : an optional :obj:`float` for the max latitude of the
          spatial coverage

        - ``creator_email``: an optional :obj:`str` the email address of the
          person who generated the IDF file

        - ``station_id`` : an optional :obj:`str` for the unique identifier of
          the station that provided the measurements contained in the granule

        - ``platform`` : an optional :obj:`str` for the name or identifier of
          the platform which hosts the sensors that provided the measurements

        - ``sensor`` : an optional :obj:`str` for the name or identifier of the
          sensor that measured the values contained in the granule

    Returns
    -------
    tuple(flatbuffers.builder.Builder, int)
        A tuple which contains two elements:

        - the :obj:`flatbuffers.builder.Builder` instance which has been used
          to serialize data
        - an :obj:`int` which is the address/offset of the serialized object
          in the builder buffer
    """
    if builder is None:
        builder = flatbuffers.Builder(0)

    cls = SEAScope.API.GranuleMetadata

    src_id = int(obj['sourceId'])
    col_id = int(obj['collectionId'])
    gra_id = int(obj['granuleId'])
    data_id = builder.CreateString(obj['dataId'])
    data_model = data_models.get(obj['dataModel'], None)
    start = obj['start']
    stop = obj['stop']

    uris = []
    uris_len = len(obj['uris'])
    for _uri in obj['uris']:
        _, idf_descriptor = SEAScope.types.idf_descriptor.serialize(builder,
                                                                    _uri)
        uris.append(idf_descriptor)
    cls.GranuleMetadataStartUrisVector(builder, uris_len)
    for i in uris[::-1]:
        builder.PrependUOffsetTRelative(i)
    _uris = builder.EndVector(uris_len)

    # optional attributes
    title = builder.CreateString(obj.get('title', ''))
    institution = builder.CreateString(obj.get('institution', ''))
    comment = builder.CreateString(obj.get('comment', ''))
    file_id = builder.CreateString(obj.get('file_id', ''))
    product_version = builder.CreateString(obj.get('product_version', ''))
    lat_min = obj.get('lat_min', 0.0)
    lat_max = obj.get('lat_max', 0.0)
    lon_min = obj.get('lon_min', 0.0)
    lon_max = obj.get('lon_max', 0.0)
    creator_email = builder.CreateString(obj.get('creator_email', ''))
    station_id = builder.CreateString(obj.get('station_id', ''))
    platform = builder.CreateString(obj.get('platform', ''))
    sensor = builder.CreateString(obj.get('sensor', ''))

    cls.GranuleMetadataStart(builder)
    cls.GranuleMetadataAddSourceId(builder, src_id)
    cls.GranuleMetadataAddCollectionId(builder, col_id)
    cls.GranuleMetadataAddGranuleId(builder, gra_id)
    cls.GranuleMetadataAddDataId(builder, data_id)
    cls.GranuleMetadataAddDataModel(builder, data_model)
    cls.GranuleMetadataAddStart(builder, start)
    cls.GranuleMetadataAddStop(builder, stop)
    cls.GranuleMetadataAddUris(builder, _uris)
    cls.GranuleMetadataAddHasTitle(builder, 'title' in obj)
    cls.GranuleMetadataAddHasInstitution(builder, 'institution' in obj)
    cls.GranuleMetadataAddHasComment(builder, 'comment' in obj)
    cls.GranuleMetadataAddHasFileId(builder, 'file_id' in obj)
    cls.GranuleMetadataAddHasProductVersion(builder, 'product_version' in obj)
    cls.GranuleMetadataAddHasLatMin(builder, 'lat_min' in obj)
    cls.GranuleMetadataAddHasLatMax(builder, 'lat_max' in obj)
    cls.GranuleMetadataAddHasLonMin(builder, 'lon_min' in obj)
    cls.GranuleMetadataAddHasLonMax(builder, 'lon_max' in obj)
    cls.GranuleMetadataAddHasCreatorEmail(builder, 'creator_email' in obj)
    cls.GranuleMetadataAddHasStationId(builder, 'station_id' in obj)
    cls.GranuleMetadataAddHasPlatform(builder, 'platform' in obj)
    cls.GranuleMetadataAddHasSensor(builder, 'sensor' in obj)
    cls.GranuleMetadataAddTitle(builder, title)
    cls.GranuleMetadataAddInstitution(builder, institution)
    cls.GranuleMetadataAddComment(builder, comment)
    cls.GranuleMetadataAddFileId(builder, file_id)
    cls.GranuleMetadataAddProductVersion(builder, product_version)
    cls.GranuleMetadataAddLatMin(builder, lat_min)
    cls.GranuleMetadataAddLatMax(builder, lat_max)
    cls.GranuleMetadataAddLonMin(builder, lon_min)
    cls.GranuleMetadataAddLonMax(builder, lon_max)
    cls.GranuleMetadataAddCreatorEmail(builder, creator_email)
    cls.GranuleMetadataAddStationId(builder, station_id)
    cls.GranuleMetadataAddPlatform(builder, platform)
    cls.GranuleMetadataAddSensor(builder, sensor)
    gm = cls.GranuleMetadataEnd(builder)
    return builder, gm


def deserialize(o):
    """
    Rebuild a granule metadata from a FlatBuffers buffer.

    Parameters
    ----------
    buf : bytearray
        The buffer which contains the granule metadata object serialized with
        FlatBuffers

    Returns
    -------
    dict
        The deserialized granule metadata object as a dictionary.
    """
    result = {}
    result['sourceId'] = o.SourceId()
    result['collectionId'] = o.CollectionId()
    result['granuleId'] = o.GranuleId()
    result['dataId'] = o.DataId()
    data_model = o.DataModel()
    for k, v in data_models.items():
        if v == data_model:
            result['dataModel'] = k
            break
    result['start'] = o.Start()
    result['stop'] = o.Stop()
    result['uris'] = []
    uris_count = o.UrisLength()
    for i in range(0, uris_count):
        u = SEAScope.types.idf_descriptor.deserialize(o.Uris(i))
        result['uris'].append(u)

    if o.HasTitle():
        result['title'] = o.Title()
    if o.HasInstitution():
        result['institution'] = o.Institution()
    if o.HasComment():
        result['comment'] = o.Comment()
    if o.HasFileId():
        result['file_id'] = o.FileId()
    if o.HasProductVersion():
        result['product_version'] = o.ProductVersion()
    if o.HasLatMin():
        result['lat_min'] = o.LatMin()
    if o.HasLatMax():
        result['lat_max'] = o.LatMax()
    if o.HasLonMin():
        result['lon_min'] = o.LonMin()
    if o.HasLonMax():
        result['lon_max'] = o.LonMax()
    if o.HasCreatorEmail():
        result['creator_email'] = o.CreatorEmail()
    if o.HasStationId():
        result['station_id'] = o.StationId()
    if o.HasPlatform():
        result['platform'] = o.Platform()
    if o.HasSensor():
        result['sensor'] = o.Sensor()
    return result



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: API

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class FloatDataBucket(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsFloatDataBucket(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = FloatDataBucket()
        x.Init(buf, n + offset)
        return x

    # FloatDataBucket
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # FloatDataBucket
    def Info(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from SEAScope.API.DataInfo import DataInfo
            obj = DataInfo()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # FloatDataBucket
    def Buffer(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Float32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # FloatDataBucket
    def BufferAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Float32Flags, o)
        return 0

    # FloatDataBucket
    def BufferLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # FloatDataBucket
    def BufferIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

def FloatDataBucketStart(builder): builder.StartObject(2)
def FloatDataBucketAddInfo(builder, info): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(info), 0)
def FloatDataBucketAddBuffer(builder, buffer): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(buffer), 0)
def FloatDataBucketStartBufferVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def FloatDataBucketEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: API

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Granule(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsGranule(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Granule()
        x.Init(buf, n + offset)
        return x

    # Granule
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Granule
    def Id(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = o + self._tab.Pos
            from SEAScope.API.RenderableId import RenderableId
            obj = RenderableId()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Granule
    def Uri(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Granule
    def CollectionId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # Granule
    def Start(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Granule
    def Stop(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Granule
    def Tags(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from SEAScope.API.Tag import Tag
            obj = Tag()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Granule
    def TagsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Granule
    def TagsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        return o == 0

def GranuleStart(builder): builder.StartObject(6)
def GranuleAddId(builder, id): builder.PrependStructSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(id), 0)
def GranuleAddUri(builder, uri): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(uri), 0)
def GranuleAddCollectionId(builder, collectionId): builder.PrependUint32Slot(2, collectionId, 0)
def GranuleAddStart(builder, start): builder.PrependUint64Slot(3, start, 0)
def GranuleAddStop(builder, stop): builder.PrependUint64Slot(4, stop, 0)
def GranuleAddTags(builder, tags): builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(tags), 0)
def GranuleStartTagsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def GranuleEnd(builder): return builder.EndObject()



# vim: ts=4:sts=4:sw=4
#
# @author: <sylvain.herledan@oceandatalab.com>
# @date: 2016-11-02
#
# This file is part of SEAScope, a 3D visualisation and analysis application
# for satellite, in-situ and numerical model data.
#
# Copyright (C) 2014-2023 OceanDataLab
#
# SEAScope is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SEAScope. If not, see <https://www.gnu.org/licenses/>.

"""
This module handles the serialization of data bucket objects
"""

import logging
import flatbuffers
import SEAScope.API.DataBucket
import SEAScope.types.data_info

logger = logging.getLogger(__name__)


def serialize(builder, bucket_obj):
    """
    Serialize a data bucket using FlatBuffers.

    Parameters
    ----------
    builder : flatbuffers.builder.Builder
        The FlatBuffers builder instance which serializes data. If this
        parameter is None, then a new builder will be created
    bucket_obj : :obj:`dict`
        Dictionary which contains information about the data bucket to
        serialize.
        It must have 2 keys:

        - ``info`` : a :obj:`dict` providing information about the data
          contained in the bucket. It must have all the keys required by
          :func:`SEAScope.types.data_info.serialize`

        - ``data`` : a :obj:`list` of `numpy.uint8` which contains values
          packed as unsigned bytes

    Returns
    -------
    tuple(flatbuffers.builder.Builder, int)
        A tuple which contains two elements:

        - the :obj:`flatbuffers.builder.Builder` instance which has been used
          to serialize data
        - an :obj:`int` which is the address/offset of the serialized object
          in the builder buffer
    """
    if builder is None:
        builder = flatbuffers.Builder(0)

    _, info = SEAScope.types.data_info.serialize(builder, bucket_obj['info'])

    cls = SEAScope.API.DataBucket

    buffer_len = len(bucket_obj['buffer'])
    cls.DataBucketStartBufferVector(builder, buffer_len)
    if '<' == bucket_obj['buffer'].dtype.str[0]:
        little_endian_data = bucket_obj['buffer']
    else:
        little_endian_data = bucket_obj['buffer'].byteswap(inplace=False)

    uoffset_t = flatbuffers.number_types.UOffsetTFlags

    # Compute how much memory is required (including alignment)
    _data_size = little_endian_data.itemsize * little_endian_data.size
    data_size = uoffset_t.py_type(_data_size)

    # Shift head
    builder.head = uoffset_t.py_type(builder.Head() - data_size)

    # Write data
    bytes_data = little_endian_data.tobytes(order='C')
    builder.Bytes[builder.Head():builder.Head() + data_size] = bytes_data
    data_buffer = builder.EndVector(buffer_len)

    cls.DataBucketStart(builder)
    cls.DataBucketAddInfo(builder, info)
    cls.DataBucketAddBuffer(builder, data_buffer)
    bucket = cls.DataBucketEnd(builder)

    return builder, bucket


def deserialize(o):
    """Not implemented"""
    pass



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: API

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class AddVariable(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsAddVariable(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = AddVariable()
        x.Init(buf, n + offset)
        return x

    # AddVariable
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # AddVariable
    def CollectionId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = o + self._tab.Pos
            from SEAScope.API.RenderableId import RenderableId
            obj = RenderableId()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # AddVariable
    def Variable(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from SEAScope.API.Variable import Variable
            obj = Variable()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

def AddVariableStart(builder): builder.StartObject(2)
def AddVariableAddCollectionId(builder, collectionId): builder.PrependStructSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(collectionId), 0)
def AddVariableAddVariable(builder, variable): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(variable), 0)
def AddVariableEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: API

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Timespan(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsTimespan(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Timespan()
        x.Init(buf, n + offset)
        return x

    # Timespan
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Timespan
    def Label(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Timespan
    def PastOffset(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Timespan
    def FutureOffset(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

def TimespanStart(builder): builder.StartObject(3)
def TimespanAddLabel(builder, label): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(label), 0)
def TimespanAddPastOffset(builder, pastOffset): builder.PrependUint64Slot(1, pastOffset, 0)
def TimespanAddFutureOffset(builder, futureOffset): builder.PrependUint64Slot(2, futureOffset, 0)
def TimespanEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: API

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class IDFDescriptor(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsIDFDescriptor(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = IDFDescriptor()
        x.Init(buf, n + offset)
        return x

    # IDFDescriptor
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # IDFDescriptor
    def Uri(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # IDFDescriptor
    def Resolution(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # IDFDescriptor
    def XArity(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # IDFDescriptor
    def YArity(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # IDFDescriptor
    def Shape(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from SEAScope.API.Shape import Shape
            obj = Shape()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # IDFDescriptor
    def SubsamplingFactor(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

def IDFDescriptorStart(builder): builder.StartObject(6)
def IDFDescriptorAddUri(builder, uri): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(uri), 0)
def IDFDescriptorAddResolution(builder, resolution): builder.PrependFloat32Slot(1, resolution, 0.0)
def IDFDescriptorAddXArity(builder, xArity): builder.PrependUint32Slot(2, xArity, 0)
def IDFDescriptorAddYArity(builder, yArity): builder.PrependUint32Slot(3, yArity, 0)
def IDFDescriptorAddShape(builder, shape): builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(shape), 0)
def IDFDescriptorAddSubsamplingFactor(builder, subsamplingFactor): builder.PrependUint32Slot(5, subsamplingFactor, 0)
def IDFDescriptorEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: API

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Collection(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsCollection(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Collection()
        x.Init(buf, n + offset)
        return x

    # Collection
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Collection
    def Id(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = o + self._tab.Pos
            from SEAScope.API.RenderableId import RenderableId
            obj = RenderableId()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Collection
    def MustBeCurrent(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # Collection
    def XSeamless(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # Collection
    def YSeamless(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # Collection
    def NewsAligned(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # Collection
    def Label(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Collection
    def Tags(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from SEAScope.API.Tag import Tag
            obj = Tag()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Collection
    def TagsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Collection
    def TagsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        return o == 0

    # Collection
    def Variables(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from SEAScope.API.Variable import Variable
            obj = Variable()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Collection
    def VariablesLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Collection
    def VariablesIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        return o == 0

    # Collection
    def DefaultVariable(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

def CollectionStart(builder): builder.StartObject(9)
def CollectionAddId(builder, id): builder.PrependStructSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(id), 0)
def CollectionAddMustBeCurrent(builder, mustBeCurrent): builder.PrependBoolSlot(1, mustBeCurrent, 0)
def CollectionAddXSeamless(builder, xSeamless): builder.PrependBoolSlot(2, xSeamless, 0)
def CollectionAddYSeamless(builder, ySeamless): builder.PrependBoolSlot(3, ySeamless, 0)
def CollectionAddNewsAligned(builder, newsAligned): builder.PrependBoolSlot(4, newsAligned, 0)
def CollectionAddLabel(builder, label): builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(label), 0)
def CollectionAddTags(builder, tags): builder.PrependUOffsetTRelativeSlot(6, flatbuffers.number_types.UOffsetTFlags.py_type(tags), 0)
def CollectionStartTagsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def CollectionAddVariables(builder, variables): builder.PrependUOffsetTRelativeSlot(7, flatbuffers.number_types.UOffsetTFlags.py_type(variables), 0)
def CollectionStartVariablesVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def CollectionAddDefaultVariable(builder, defaultVariable): builder.PrependUint32Slot(8, defaultVariable, 0)
def CollectionEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: API

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class SearchGranules(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsSearchGranules(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = SearchGranules()
        x.Init(buf, n + offset)
        return x

    # SearchGranules
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # SearchGranules
    def Current(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # SearchGranules
    def Start(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # SearchGranules
    def Stop(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # SearchGranules
    def Zoom(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # SearchGranules
    def Collections(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 8
            from SEAScope.API.CollectionId import CollectionId
            obj = CollectionId()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # SearchGranules
    def CollectionsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # SearchGranules
    def CollectionsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        return o == 0

def SearchGranulesStart(builder): builder.StartObject(5)
def SearchGranulesAddCurrent(builder, current): builder.PrependUint64Slot(0, current, 0)
def SearchGranulesAddStart(builder, start): builder.PrependUint64Slot(1, start, 0)
def SearchGranulesAddStop(builder, stop): builder.PrependUint64Slot(2, stop, 0)
def SearchGranulesAddZoom(builder, zoom): builder.PrependFloat32Slot(3, zoom, 0.0)
def SearchGranulesAddCollections(builder, collections): builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(collections), 0)
def SearchGranulesStartCollectionsVector(builder, numElems): return builder.StartVector(8, numElems, 4)
def SearchGranulesEnd(builder): return builder.EndObject()



# vim: ts=4:sts=4:sw=4
#
# @author: <sylvain.herledan@oceandatalab.com>
# @date: 2016-09-09
#
# This file is part of SEAScope, a 3D visualisation and analysis application
# for satellite, in-situ and numerical model data.
#
# Copyright (C) 2014-2023 OceanDataLab
#
# SEAScope is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SEAScope. If not, see <https://www.gnu.org/licenses/>.

"""
This module handles the serialization and deserialization of GCP objects
"""

import logging
import flatbuffers
import SEAScope.API.GCP

logger = logging.getLogger(__name__)


def serialize(builder, gcp_obj):
    """
    Serialize a GCP using FlatBuffers.

    Parameters
    ----------
    builder : flatbuffers.builder.Builder
        The FlatBuffers builder instance which serializes data. If this
        parameter is None, then a new builder will be created
    gcp_obj : dict
        Dictionary which contains information about the GCP to serialize

        It must have four keys:

        - ``lon`` : a :obj:`float` which gives the longitude of the GCP

        - ``lat`` : a :obj:`float` which gives the latitude of the GCP

        - ``i`` : an :obj:`int` which is the number of the column in the data
          matrix for the GCP

        - ``j`` : an :obj:`int` which is the number of the row in the data
          matrix for the GCP

        There is a fifth optional key:

        - ``time`` : a :obj:`int` which is the number of milliseconds since
          epoch for the GCP. The default value is 0 if this key is not
          provided.

    Returns
    -------
    tuple(flatbuffers.builder.Builder, int)
        A tuple which contains two elements:

        - the :obj:`flatbuffers.builder.Builder` instance which has been used
          to serialize data
        - an :obj:`int` which is the address/offset of the serialized object
          in the builder buffer
    """
    if builder is None:
        builder = flatbuffers.Builder(0)

    time_value = gcp_obj.get('time', 0)
    cls = SEAScope.API.GCP
    gcp = cls.CreateGCP(builder,
                        time_value,
                        gcp_obj['lon'],
                        gcp_obj['lat'],
                        int(gcp_obj['i']),
                        int(gcp_obj['j']))

    return builder, gcp


def deserialize(o):
    """
    Rebuild a GCP from a FlatBuffers buffer.

    Parameters
    ----------
    buf : bytearray
        The buffer which contain the GCP object serialized with FlatBuffers

    Returns
    -------
    dict
        The deserialized GCP object as a dictionary.
    """
    result = {}
    result['time'] = o.Time()
    result['lon'] = o.Lon()
    result['lat'] = o.Lat()
    result['i'] = o.I()
    result['j'] = o.J()
    return result



# vim: ts=4:sts=4:sw=4
#
# @author: <sylvain.herledan@oceandatalab.com>
# @date: 2016-11-02
#
# This file is part of SEAScope, a 3D visualisation and analysis application
# for satellite, in-situ and numerical model data.
#
# Copyright (C) 2014-2023 OceanDataLab
#
# SEAScope is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SEAScope. If not, see <https://www.gnu.org/licenses/>.

"""
This module handles the serialization of data info objects
"""

import logging
import flatbuffers
import SEAScope.API.DataInfo
import SEAScope.API.DataType

logger = logging.getLogger(__name__)
data_types = {
    'none': SEAScope.API.DataType.DataType.NONE,
    'uint8': SEAScope.API.DataType.DataType.UBYTE,
    'ushort16': SEAScope.API.DataType.DataType.USHORT,
    'uint32': SEAScope.API.DataType.DataType.UINT,
    'float32': SEAScope.API.DataType.DataType.FLOAT32}
"""
dict: Data types that can be used to store information in SEAScope. Beware:
SEAScope expects some values to have a specific type and will probably crash
or return an error if a value with another type is provided.
"""


def serialize(builder, info_obj):
    """
    Serialize a granule data info using FlatBuffers.

    Parameters
    ----------
    builder : flatbuffers.builder.Builder
        The FlatBuffers builder instance which serializes data. If this
        parameter is None, then a new builder will be created
    info_obj : dict
        Dictionary which contains information about the data
        to serialize

        It must have  keys:

        - ``offsets`` : a :obj:`list` of :obj:`float` for the offset used when
          packing the data values (``packed = (values - offset) / scale``).

          One value per channel contained in the data matrix

        - ``scaleFactors`` : a :obj:`list` of `float` for the scale factors
          used when packing the data values (``packed = (values - offset) /
          scale``)

          One value per channel contained in the data matrix

        - ``fillValues`` : a :obj:`list` of :obj:`float` for the fill values
          that have been stored in the data matrix to replace missing/invalid
          data. Note that the fill values in this list belong to the packed
          values domain.

          One value per channel contained in the data matrix

        - ``dataType`` : a :obj:`str` which identifies the type to allocate for
          the values contained in the data matrix. Its value must be a key of
          the :const:`SEAScope.types.data_info.data_types`` dictionary.

          Only ``uint8`` is supported at the moment

        - ``channels`` : an :obj:`int` for the number of channels/fields
          contained in the data matrix

        - ``xArity`` : an :obj:`int` which tells SEAScope how many columns it
          must allocate when reconstructing the data matrix

        - ``yArity`` : an :obj:`int` which tells SEAScope how many rows it
          must allocate when reconstructing the data matrix

    Returns
    -------
    tuple(flatbuffers.builder.Builder, int)
        A tuple which contains two elements:

        - the :obj:`flatbuffers.builder.Builder` instance which has been used
          to serialize data
        - an :obj:`int` which is the address/offset of the serialized object
          in the builder buffer
    """
    if builder is None:
        builder = flatbuffers.Builder(0)

    cls = SEAScope.API.DataInfo

    offsets_len = len(info_obj['offsets'])
    cls.DataInfoStartOffsetsVector(builder, offsets_len)
    for v in info_obj['offsets'][::-1]:
        builder.PrependFloat32(v)
    offsets = builder.EndVector(offsets_len)

    scales_len = len(info_obj['scaleFactors'])
    cls.DataInfoStartScaleFactorsVector(builder, scales_len)
    for v in info_obj['scaleFactors'][::-1]:
        builder.PrependFloat32(v)
    scale_factors = builder.EndVector(scales_len)

    fills_len = len(info_obj['fillValues'])
    cls.DataInfoStartFillValuesVector(builder, fills_len)
    for v in info_obj['fillValues'][::-1]:
        builder.PrependFloat32(v)
    fill_values = builder.EndVector(fills_len)

    _data_type = data_types.get(info_obj['dataType'], data_types['none'])

    cls.DataInfoStart(builder)
    cls.DataInfoAddChannels(builder, int(info_obj['channels']))
    cls.DataInfoAddXArity(builder, int(info_obj['xArity']))
    cls.DataInfoAddYArity(builder, int(info_obj['yArity']))
    cls.DataInfoAddDataType(builder, _data_type)
    cls.DataInfoAddOffsets(builder, offsets)
    cls.DataInfoAddScaleFactors(builder, scale_factors)
    cls.DataInfoAddFillValues(builder, fill_values)
    data_info = cls.DataInfoEnd(builder)

    return builder, data_info


def deserialize(o):
    """
    Not implemented
    """
    pass



# vim: ts=4:sts=4:sw=4
#
# @author: <sylvain.herledan@oceandatalab.com>
# @date: 2022-08-22
#
# This file is part of SEAScope, a 3D visualisation and analysis application
# for satellite, in-situ and numerical model data.
#
# Copyright (C) 2014-2023 OceanDataLab
#
# SEAScope is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SEAScope. If not, see <https://www.gnu.org/licenses/>.

"""
This module handles the serialization of the checkSchemaVersion command and
the deserialization of the result sent by SEAScope
"""

import math
import logging
import flatbuffers
import SEAScope.API.OpCode
import SEAScope.API.CheckSchemaVersion
import SEAScope.API.CheckSchemaVersionResponse
import SEAScope.API.Command

logger = logging.getLogger(__name__)


def _get_local_schema_version():
    # Current schema version is stored as the default value for the "Version"
    # attribute in CheckSchemaVersionResponse
    class DummyTab:
        def Offset(_):
            return 0
    response_module = SEAScope.API.CheckSchemaVersionResponse
    dummy_response = response_module.CheckSchemaVersionResponse()
    dummy_response._tab = DummyTab
    return dummy_response.Version()


def serialize(builder):
    """
    Serialize a checkBindingsVersion command using FlatBuffers.

    Parameters
    ----------
    builder : flatbuffers.builder.Builder
        The FlatBuffers builder instance which serializes data. If this
        parameter is None, then a new builder will be created

    Returns
    -------
    tuple(flatbuffers.builder.Builder, int)
        A tuple which contains two elements:

        - the :obj:`flatbuffers.builder.Builder` instance which has been used
          to serialize data
        - an :obj:`int` which is the address/offset of the serialized object
          in the builder buffer
    """
    if builder is None:
        builder = flatbuffers.Builder(0)

    local_schema_version = _get_local_schema_version()

    cls = SEAScope.API.CheckSchemaVersion
    cls.CheckSchemaVersionStart(builder)
    cls.CheckSchemaVersionAddVersion(builder, local_schema_version)
    cmd_args = cls.CheckSchemaVersionEnd(builder)

    opcode = SEAScope.API.OpCode.OpCode().checkSchemaVersion
    args_type = SEAScope.API.CommandArgs.CommandArgs().CheckSchemaVersion
    SEAScope.API.Command.CommandStart(builder)
    SEAScope.API.Command.CommandAddOpcode(builder, opcode)
    SEAScope.API.Command.CommandAddArgsType(builder, args_type)
    SEAScope.API.Command.CommandAddArgs(builder, cmd_args)
    cmd = SEAScope.API.Command.CommandEnd(builder)

    return builder, cmd


def deserialize(buf):
    """
    Deserialize the response that SEAScope sends after executing the
    checkSchemaVersion command

    Parameters
    ----------
    buf : bytearray
        The buffer which contains the result of the checkSchemaVersion
        command serialized with FlatBuffers

    Returns
    -------
    tuple(bool, str)
        A tuple which contains two elements:

        - a :obj:`bool` which tells if the schema versions match
        - a :obj:`str` that contains the version of the latest SEAScope Python
          package which has full compatibility with the viewer
    """
    cls = SEAScope.API.CheckSchemaVersionResponse.CheckSchemaVersionResponse
    res = cls.GetRootAsCheckSchemaVersionResponse(buf, 0)

    local_schema_version = _get_local_schema_version()
    remote_schema_version = math.floor(res.Version() / 203)
    version_match = (local_schema_version == remote_schema_version)

    major = math.floor(res.PythonVersionMajor() / 203)
    minor = math.floor(res.PythonVersionMinor() / 203)
    build = math.floor(res.PythonVersionBuild() / 203)
    recommended_python_version = f'{major}.{minor}.{build}'

    return version_match, recommended_python_version



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: API

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class GCP(object):
    __slots__ = ['_tab']

    # GCP
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # GCP
    def Time(self): return self._tab.Get(flatbuffers.number_types.Uint64Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(0))
    # GCP
    def Lon(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(8))
    # GCP
    def Lat(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(12))
    # GCP
    def I(self): return self._tab.Get(flatbuffers.number_types.Uint32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(16))
    # GCP
    def J(self): return self._tab.Get(flatbuffers.number_types.Uint32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(20))

def CreateGCP(builder, time, lon, lat, i, j):
    builder.Prep(8, 24)
    builder.PrependUint32(j)
    builder.PrependUint32(i)
    builder.PrependFloat32(lat)
    builder.PrependFloat32(lon)
    builder.PrependUint64(time)
    return builder.Offset()



# vim: ts=4:sts=4:sw=4
#
# @author: <sylvain.herledan@oceandatalab.com>
# @date: 2016-09-08
#
# This file is part of SEAScope, a 3D visualisation and analysis application
# for satellite, in-situ and numerical model data.
#
# Copyright (C) 2014-2023 OceanDataLab
#
# SEAScope is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SEAScope. If not, see <https://www.gnu.org/licenses/>.

"""
This module handles the serialization of the searchGranules command and the
deserialization of the results sent by SEAScope
"""

import logging
import flatbuffers
import SEAScope.API.OpCode
import SEAScope.API.SearchGranules
import SEAScope.API.SearchGranulesResponse
import SEAScope.API.Command
import SEAScope.types.collection_id
import SEAScope.types.granule_metadata

logger = logging.getLogger(__name__)


def serialize(builder, search_obj):
    """
    Serialize a searchGranules command using FlatBuffers.

    Parameters
    ----------
    builder : flatbuffers.builder.Builder
        The FlatBuffers builder instance which serializes data. If this
        parameter is None, then a new builder will be created
    search_obj : dict

        - collections : a :obj:`list` of :obj:`dict` which tells SEAScope the
          collections it must consider during the search. The dict values must
          satisfy the requirements of the
          :func:`SEAScope.types.collection_id.serialize` method

        - current : an :obj:`int` which acts in the same way as the current
          datetime of the timeline. Expressed in seconds since
          1970-01-01T00:00:00Z

          If a collection has the ``mustBeCurrent`` flag set to True, then
          SEAScope will only return the granules for this collection if
          ``current`` belongs to [granule.start, granule.stop[

        - start : an :obj:`int` which is the lower time bound used by SEAScope
          when searching for granules, expressed in seconds since
          1970-01-01T00:00:00Z.

          In interactive usage of SEAScope, this has the ``currentDatetime -
          pastOffset`` value

        - stop : an :obj:`int` which is the upper bound used by SEAScope when
          searching for granules, expressed in seconds since
          1970-01-01T00:00:00Z.

          In interactive usage of SEAScope, this has the ``currentDatetime +
          futureOffset`` value

        - zoom : a :obj:`float` that tells SEAScope at which altitude the
          camera should be during the search. It has an effect on the
          resolution of the granules returned in the results.

    Returns
    -------
    tuple(flatbuffers.builder.Builder, int)
        A tuple which contains two elements:

        - the :obj:`flatbuffers.builder.Builder` instance which has been used
          to serialize data
        - an :obj:`int` which is the address/offset of the serialized object
          in the builder buffer
    """
    if builder is None:
        builder = flatbuffers.Builder(0)

    cls = SEAScope.API.SearchGranules

    collectionsOffsets = 0
    cls.SearchGranulesStartCollectionsVector(builder,
                                             len(search_obj['collections']))
    for collection in search_obj['collections']:
        SEAScope.types.collection_id.serialize(builder, collection)
    collectionsOffsets = builder.EndVector(len(search_obj['collections']))

    cls.SearchGranulesStart(builder)
    cls.SearchGranulesAddCurrent(builder, search_obj['current'])
    cls.SearchGranulesAddStart(builder, search_obj['start'])
    cls.SearchGranulesAddStop(builder, search_obj['stop'])
    cls.SearchGranulesAddZoom(builder, search_obj['zoom'])
    cls.SearchGranulesAddCollections(builder, collectionsOffsets)
    cmd_args = cls.SearchGranulesEnd(builder)

    opcode = SEAScope.API.OpCode.OpCode().searchGranules
    args_type = SEAScope.API.CommandArgs.CommandArgs().SearchGranules
    SEAScope.API.Command.CommandStart(builder)
    SEAScope.API.Command.CommandAddOpcode(builder, opcode)
    SEAScope.API.Command.CommandAddArgsType(builder, args_type)
    SEAScope.API.Command.CommandAddArgs(builder, cmd_args)
    cmd = SEAScope.API.Command.CommandEnd(builder)

    return builder, cmd


def deserialize(buf):
    """
    Deserialize the response that SEAScope sends after executing the
    searchGranules command

    Parameters
    ----------
    buf : bytearray
        The buffer which contains the result of the searchGranules command
        serialized with FlatBuffers

    Returns
    -------
    list of dict
        The metadata of the granules found by SEAScope. Each dict is the
        result of the :func:`SEAScope.types.granule_metadata.deserialize`
        method (see source code for more details)
    """
    deserializer = SEAScope.types.granule_metadata.deserialize
    cls = SEAScope.API.SearchGranulesResponse.SearchGranulesResponse
    res = cls.GetRootAsSearchGranulesResponse(buf, 0)

    granules_len = res.GranulesLength()
    return [deserializer(res.Granules(x)) for x in range(0, granules_len)]



# vim: ts=4:sts=4:sw=4
#
# @author: <sylvain.herledan@oceandatalab.com>
# @date: 2016-09-07
#
# This file is part of SEAScope, a 3D visualisation and analysis application
# for satellite, in-situ and numerical model data.
#
# Copyright (C) 2014-2023 OceanDataLab
#
# SEAScope is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SEAScope. If not, see <https://www.gnu.org/licenses/>.

"""
This module handles the serialization and deserialization of the identifiers
for renderable objects.
"""

import logging
import flatbuffers
import SEAScope.API.RenderableId

logger = logging.getLogger(__name__)


def serialize(builder, rid_obj):
    """
    Serialize a renderable ID using FlatBuffers.

    Parameters
    ----------
    builder : flatbuffers.builder.Builder
        The FlatBuffers builder instance which serializes data. If this
        parameter is None, then a new builder will be created
    rid_obj : dict
        Dictionary which contains information about the renderable ID to
        serialize

        It must have 5 keys:

        - ``granuleLevel`` : a :obj:`bool` that tells SEAScope if the targeted
          object is a granule (True) or a collection (False)

        - ``sourceId`` : a :obj:`int` which is the unique identifier of the
          data source which provided the targeted object

        - ``collectionId`` : a :obj:`int` which is the unique identifier
          (within the data source) of the targeted collection

        - ``granuleId`` : a :obj:`int` which is the unique identifier (within
          the data source) of the targeted granule

          Ignored if ``granuleLevel`` is False

        - ``variableId`` :a :obj:`int` which is the unique identifier (within
          the collection) of the targeted variable

    Returns
    -------
    tuple(flatbuffers.builder.Builder, int)
        A tuple which contains two elements:

        - the :obj:`flatbuffers.builder.Builder` instance which has been used
          to serialize data
        - an :obj:`int` which is the address/offset of the serialized object
          in the builder buffer
    """
    if builder is None:
        builder = flatbuffers.Builder(0)

    cls = SEAScope.API.RenderableId
    rid = cls.CreateRenderableId(builder,
                                 int(rid_obj['granuleLevel']),
                                 int(rid_obj['sourceId']),
                                 int(rid_obj['collectionId']),
                                 int(rid_obj['granuleId']),
                                 int(rid_obj['variableId']))

    return builder, rid


def deserialize(o):
    """
    Rebuild an identifier for a renderable object from a FlatBuffers buffer.

    Parameters
    ----------
    buf : bytearray
        The buffer which contain the renderable ID object serialized with
        FlatBuffers

    Returns
    -------
    dict
        The deserialized renderable ID object as a dictionary.
    """
    result = {}
    result['granuleLevel'] = 0 < o.GranuleLevel()
    result['sourceId'] = o.SourceId()
    result['collectionId'] = o.CollectionId()
    result['granuleId'] = o.GranuleId()
    result['variableId'] = o.VariableId()
    return result



# vim: ts=4:sts=4:sw=4
#
# @author: <sylvain.herledan@oceandatalab.com>
# @date: 2016-10-20
#
# This file is part of SEAScope, a 3D visualisation and analysis application
# for satellite, in-situ and numerical model data.
#
# Copyright (C) 2014-2023 OceanDataLab
#
# SEAScope is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SEAScope. If not, see <https://www.gnu.org/licenses/>.

"""
This module handles the serialization of the getExtraction command and the
deserialization of the result sent by SEAScope
"""

import logging
import flatbuffers
import SEAScope.API.OpCode
import SEAScope.API.GetExtractionResponse
import SEAScope.API.Command
import SEAScope.types.extraction

logger = logging.getLogger(__name__)


def serialize(builder):
    """
    Serialize a getExtraction command using FlatBuffers.

    Parameters
    ----------
    builder : flatbuffers.builder.Builder
        The FlatBuffers builder instance which serializes data. If this
        parameter is None, then a new builder will be created

    Returns
    -------
    tuple(flatbuffers.builder.Builder, int)
        A tuple which contains two elements:

        - the :obj:`flatbuffers.builder.Builder` instance which has been used
          to serialize data
        - an :obj:`int` which is the address/offset of the serialized object
          in the builder buffer
    """
    if builder is None:
        builder = flatbuffers.Builder(0)

    opcode = SEAScope.API.OpCode.OpCode().getExtraction
    SEAScope.API.Command.CommandStart(builder)
    SEAScope.API.Command.CommandAddOpcode(builder, opcode)
    cmd = SEAScope.API.Command.CommandEnd(builder)

    return builder, cmd


def deserialize(buf):
    """
    Deserialize the response that SEAScope sends after executing the
    getExtraction command

    Parameters
    ----------
    buf : bytearray
        The buffer which contains the result of the getExtraction command
        serialized with FlatBuffers

    Returns
    -------
    list of dict
        The result contains the data extracted by SEAScope from every granule
        currently rendered as a raster or as a trajectory.

        Each dict is the result of the
        :func:`SEAScope.types.extraction.deserialize` method (see source code
        for more details)
    """
    cls = SEAScope.API.GetExtractionResponse.GetExtractionResponse
    res = cls.GetRootAsGetExtractionResponse(buf, 0)

    extractions_count = res.ExtractionsLength()
    result = []
    for i in range(0, extractions_count):
        extraction = SEAScope.types.extraction.deserialize(res.Extractions(i))
        result.append(extraction)
    return result



# vim: ts=4:sts=4:sw=4
#
# @author: <sylvain.herledan@oceandatalab.com>
# @date: 2017-05-15
#
# This file is part of SEAScope, a 3D visualisation and analysis application
# for satellite, in-situ and numerical model data.
#
# Copyright (C) 2014-2023 OceanDataLab
#
# SEAScope is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SEAScope. If not, see <https://www.gnu.org/licenses/>.

"""
This module handles the serialization of the resetCamera command
"""

import logging
import flatbuffers
import SEAScope.API.OpCode
import SEAScope.API.Command

logger = logging.getLogger(__name__)


def serialize(builder):
    """
    Serialize a resetCamera command using FlatBuffers.

    Parameters
    ----------
    builder : flatbuffers.builder.Builder
        The FlatBuffers builder instance which serializes data. If this
        parameter is None, then a new builder will be created

    Returns
    -------
    tuple(flatbuffers.builder.Builder, int)
        A tuple which contains two elements:

        - the :obj:`flatbuffers.builder.Builder` instance which has been used
          to serialize data
        - an :obj:`int` which is the address/offset of the serialized object
          in the builder buffer
    """
    if builder is None:
        builder = flatbuffers.Builder(0)

    opcode = SEAScope.API.OpCode.OpCode().resetCamera
    SEAScope.API.Command.CommandStart(builder)
    SEAScope.API.Command.CommandAddOpcode(builder, opcode)
    cmd = SEAScope.API.Command.CommandEnd(builder)

    return builder, cmd


def deserialize():
    """
    Not implemented
    """
    pass



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: API

class CommandArgs(object):
    NONE = 0
    AddSource = 1
    RemoveSource = 2
    AddCollection = 3
    RemoveCollection = 4
    AddVariable = 5
    RemoveVariable = 6
    SetCurrentDatetime = 7
    AddTimespan = 8
    RemoveTimespan = 9
    SetTimespan = 10
    AddTimestep = 11
    RemoveTimestep = 12
    SetTimestep = 13
    SetNearestMode = 14
    SetDataAlwaysVisibleMode = 15
    SearchGranules = 16
    SetRendering = 17
    GetRendering = 18
    AddGranule = 19
    LookAt = 20
    SetAltitude = 21
    SelectVariable = 22
    SelectVariableByLabel = 23
    GetVariableIdentifier = 24
    CheckSchemaVersion = 25




# vim: set ts=4:sts=4:sw=4
#
# @author: <sylvain.herledan@oceandatalab.com>
#
# This file is part of SEAScope, a 3D visualisation and analysis application
# for satellite, in-situ and numerical model data.
#
# Copyright (C) 2014-2023 OceanDataLab
#
# SEAScope is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SEAScope. If not, see <https://www.gnu.org/licenses/>.

"""
This module provides some helper methods required by the seascope-processor
command to process data extracted by SEAScope.
"""

import struct
import numpy
import logging

import SEAScope.cmds.get_extraction
import SEAScope.lib.utils
import SEAScope.upload
try:
    import cPickle as pickle
except ImportError:
    import pickle

logger = logging.getLogger(__name__)


def get_dists(mask):
    """Compute the cumulative distance between a list of locations provided
    by SEAScope when the user makes a transect.

    Parameters
    ----------
    mask : dict
        Data structure (similar to granules) created by SEAscope when the user
        extracts data along a polyline (transect)

    Returns
    -------
    list of float
        The cumulative distances covered at each point while following the
        transect, in kilometers
    """
    import pyproj
    geod = pyproj.Geod(ellps='WGS84')
    gcps = mask['meta']['gcps']

    dists = [0.0]
    last_lon = gcps[0]['lon']
    last_lat = gcps[0]['lat']
    logger.debug('Number of steps: {}'.format(len(gcps)))

    for n in range(1, len(gcps)):
        lon = gcps[n]['lon']
        lat = gcps[n]['lat']
        _, _, d = geod.inv(last_lon, last_lat, lon, lat)
        dists.append(dists[-1] + .001 * d)
        last_lon = lon
        last_lat = lat

    return dists


def _arrays2rgb(channels):
    """Deprecated?"""
    ubyte_data = []
    for n in range(0, 3):
        d_min = numpy.min(channels[n])
        d_max = numpy.max(channels[n])
        extent = d_min - d_max
        if d_min == d_max:
            extent = 1.0
        d_ubyte = 255 * (channels[n] - d_min) / extent
        ubyte_data.append(d_ubyte)
    rgb = numpy.stack(ubyte_data, axis=2).astype(numpy.ubyte)
    return rgb


def _read_response(link, msg_size):
    """Fetch a fixed number of bytes from the socket connected to SEAScope.

    Parameters
    ----------
    link : socket.socket
        Stream socket connected to SEAScope
    msg_size : int
        Number of bytes to fetch from the socket

    Returns
    -------
    :obj:`list` of :obj:`bytes`
        Raw data sent by SEAScope"""
    chunk_size = msg_size
    already_read = 0
    result = b''
    while already_read < msg_size:
        buf = link.recv(chunk_size)
        already_read += len(buf)
        result += buf
        chunk_size = msg_size - already_read

    return result


def get_extracted_data(host='localhost', port=11155):
    """Fetch extracted data from SEAScope and format the result to make it
    easier to manipulate.

    Parameters
    ----------
    host : str
        IP address of the network interface that the SEAScope application
        listens to.
    port : int
        Port number that the SEAScope application listens to.

    Returns
    -------
    dict
        The data extracted by SEAScope, reformatted to be easier to analyse and
        manipulate
    """
    logger.info('Retrieving extracted data...')
    with SEAScope.upload.connect(host, port) as s:
        serializer = SEAScope.cmds.get_extraction.serialize
        builder = None
        builder, serialized = serializer(builder)
        builder.Finish(serialized)
        buf = builder.Output()
        s.sendall(struct.pack('>Q', len(buf))+buf)

        buf = s.recv(8)
        msg_len = struct.unpack('<Q', buf)
        logger.debug('Read {} bytes'.format(msg_len[0]))
        buf = _read_response(s, msg_len[0])
        logger.debug('{} bytes read'.format(len(buf)))
        deserializer = SEAScope.cmds.get_extraction.deserialize
        raw_result = deserializer(buf)

    clean_data = SEAScope.lib.utils.raw2data(raw_result)
    del raw_result
    return clean_data


def save_pyo(obj, output_path):
    """Serialize a Python object and save the result on the filesystem.

    Parameters
    ----------
    obj : `object`
        A serializable Python object
    output_path : str
        Path of the file that will contain the serialized object
    """
    with open(output_path, 'wb') as f:
        pickle.dump(obj, f, protocol=1)


def load_pyo(input_path):
    """Load a file from the filesystem and deserialize its content as a Python
    object.

    Parameters
    ----------
    input_path : str
        Path of the file which contains a serialized Python object

    Returns
    -------
    object
        The deserialized Python object
    """
    obj = None
    with open(input_path, 'rb') as f:
        obj = pickle.load(f)
    return obj



# vim: ts=4:sts=4:sw=4
#
# @author: <sylvain.herledan@oceandatalab.com>
# @date: 2016-09-07
#
# This file is part of SEAScope, a 3D visualisation and analysis application
# for satellite, in-situ and numerical model data.
#
# Copyright (C) 2014-2023 OceanDataLab
#
# SEAScope is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SEAScope. If not, see <https://www.gnu.org/licenses/>.

"""
This module handles the serialization and deserialization of source objects
"""

import logging
import flatbuffers
import SEAScope.API.Source
import SEAScope.types.tag

logger = logging.getLogger(__name__)


def serialize(builder, src_obj):
    """
    Serialize a source using FlatBuffers.

    Parameters
    ----------
    builder : flatbuffers.builder.Builder
        The FlatBuffers builder instance which serializes data. If this
        parameter is None, then a new builder will be created
    src_obj : dict
        Dictionary which contains information about the source to serialize
        It must have ten keys:

        - ``title`` : a :obj:`str` that SEAScope uses to designate the source

        - ``description``: a :obj:`str` that SEAScope can display to provide
          information about the source

        - ``tags`` : a :obj:`dict` which contains tags that describe the
          collection. SEAScope considers a (key, value) pair as a tag.

        - ``id`` : an :obj:`int` that SEAScope can uses as a unique identifier
          for the source

    Returns
    -------
    tuple(flatbuffers.builder.Builder, int)
        A tuple which contains two elements:

        - the :obj:`flatbuffers.builder.Builder` instance which has been used
          to serialize data
        - an :obj:`int` which is the address/offset of the serialized object
          in the builder buffer
    """
    if builder is None:
        builder = flatbuffers.Builder(0)

    cls = SEAScope.API.Source

    _title = builder.CreateString(src_obj['title'])
    _description = builder.CreateString(src_obj['description'])

    tags = []
    tags_len = len(src_obj['tags'].keys())
    for k, v in src_obj['tags'].items():
        _, tag = SEAScope.types.tag.serialize(builder, k, v)
        tags.append(tag)
    cls.SourceStartTagsVector(builder, tags_len)
    for t in tags:
        builder.PrependUOffsetTRelative(t)
    _tags = builder.EndVector(tags_len)

    cls.SourceStart(builder)
    cls.SourceAddId(builder, int(src_obj['id']))
    cls.SourceAddTitle(builder, _title)
    cls.SourceAddDescription(builder, _description)
    cls.SourceAddTags(builder, _tags)
    source = cls.SourceEnd(builder)

    return builder, source


def deserialize(o):
    """"""
    src_obj = {}
    src_obj['id'] = o.Id()
    src_obj['title'] = o.Title()
    src_obj['description'] = o.Description()
    src_obj['tags'] = {}
    tags_count = o.TagsLength()
    for i in range(0, tags_count):
        name, value = SEAScope.types.tag.deserialize(o.Tags(i))
        src_obj['tags'][name] = value

    return src_obj



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: API

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class GranuleMetadata(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsGranuleMetadata(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = GranuleMetadata()
        x.Init(buf, n + offset)
        return x

    # GranuleMetadata
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # GranuleMetadata
    def SourceId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # GranuleMetadata
    def CollectionId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # GranuleMetadata
    def GranuleId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # GranuleMetadata
    def DataId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # GranuleMetadata
    def DataModel(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # GranuleMetadata
    def Start(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # GranuleMetadata
    def Stop(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # GranuleMetadata
    def Uris(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from SEAScope.API.IDFDescriptor import IDFDescriptor
            obj = IDFDescriptor()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # GranuleMetadata
    def UrisLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # GranuleMetadata
    def UrisIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        return o == 0

    # GranuleMetadata
    def HasTitle(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # GranuleMetadata
    def HasInstitution(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # GranuleMetadata
    def HasComment(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # GranuleMetadata
    def HasFileId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # GranuleMetadata
    def HasProductVersion(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # GranuleMetadata
    def HasLatMin(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # GranuleMetadata
    def HasLatMax(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # GranuleMetadata
    def HasLonMin(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(34))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # GranuleMetadata
    def HasLonMax(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(36))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # GranuleMetadata
    def HasCreatorEmail(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(38))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # GranuleMetadata
    def HasStationId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(40))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # GranuleMetadata
    def HasPlatform(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(42))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # GranuleMetadata
    def HasSensor(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(44))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # GranuleMetadata
    def Title(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(46))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # GranuleMetadata
    def Institution(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(48))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # GranuleMetadata
    def Comment(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(50))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # GranuleMetadata
    def FileId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(52))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # GranuleMetadata
    def ProductVersion(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(54))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # GranuleMetadata
    def LatMin(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(56))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # GranuleMetadata
    def LatMax(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(58))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # GranuleMetadata
    def LonMin(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(60))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # GranuleMetadata
    def LonMax(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(62))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # GranuleMetadata
    def CreatorEmail(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(64))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # GranuleMetadata
    def StationId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(66))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # GranuleMetadata
    def Platform(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(68))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # GranuleMetadata
    def Sensor(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(70))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

def GranuleMetadataStart(builder): builder.StartObject(34)
def GranuleMetadataAddSourceId(builder, sourceId): builder.PrependUint32Slot(0, sourceId, 0)
def GranuleMetadataAddCollectionId(builder, collectionId): builder.PrependUint32Slot(1, collectionId, 0)
def GranuleMetadataAddGranuleId(builder, granuleId): builder.PrependUint64Slot(2, granuleId, 0)
def GranuleMetadataAddDataId(builder, dataId): builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(dataId), 0)
def GranuleMetadataAddDataModel(builder, dataModel): builder.PrependInt8Slot(4, dataModel, 0)
def GranuleMetadataAddStart(builder, start): builder.PrependUint64Slot(5, start, 0)
def GranuleMetadataAddStop(builder, stop): builder.PrependUint64Slot(6, stop, 0)
def GranuleMetadataAddUris(builder, uris): builder.PrependUOffsetTRelativeSlot(7, flatbuffers.number_types.UOffsetTFlags.py_type(uris), 0)
def GranuleMetadataStartUrisVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def GranuleMetadataAddHasTitle(builder, hasTitle): builder.PrependBoolSlot(8, hasTitle, 0)
def GranuleMetadataAddHasInstitution(builder, hasInstitution): builder.PrependBoolSlot(9, hasInstitution, 0)
def GranuleMetadataAddHasComment(builder, hasComment): builder.PrependBoolSlot(10, hasComment, 0)
def GranuleMetadataAddHasFileId(builder, hasFileId): builder.PrependBoolSlot(11, hasFileId, 0)
def GranuleMetadataAddHasProductVersion(builder, hasProductVersion): builder.PrependBoolSlot(12, hasProductVersion, 0)
def GranuleMetadataAddHasLatMin(builder, hasLatMin): builder.PrependBoolSlot(13, hasLatMin, 0)
def GranuleMetadataAddHasLatMax(builder, hasLatMax): builder.PrependBoolSlot(14, hasLatMax, 0)
def GranuleMetadataAddHasLonMin(builder, hasLonMin): builder.PrependBoolSlot(15, hasLonMin, 0)
def GranuleMetadataAddHasLonMax(builder, hasLonMax): builder.PrependBoolSlot(16, hasLonMax, 0)
def GranuleMetadataAddHasCreatorEmail(builder, hasCreatorEmail): builder.PrependBoolSlot(17, hasCreatorEmail, 0)
def GranuleMetadataAddHasStationId(builder, hasStationId): builder.PrependBoolSlot(18, hasStationId, 0)
def GranuleMetadataAddHasPlatform(builder, hasPlatform): builder.PrependBoolSlot(19, hasPlatform, 0)
def GranuleMetadataAddHasSensor(builder, hasSensor): builder.PrependBoolSlot(20, hasSensor, 0)
def GranuleMetadataAddTitle(builder, title): builder.PrependUOffsetTRelativeSlot(21, flatbuffers.number_types.UOffsetTFlags.py_type(title), 0)
def GranuleMetadataAddInstitution(builder, institution): builder.PrependUOffsetTRelativeSlot(22, flatbuffers.number_types.UOffsetTFlags.py_type(institution), 0)
def GranuleMetadataAddComment(builder, comment): builder.PrependUOffsetTRelativeSlot(23, flatbuffers.number_types.UOffsetTFlags.py_type(comment), 0)
def GranuleMetadataAddFileId(builder, fileId): builder.PrependUOffsetTRelativeSlot(24, flatbuffers.number_types.UOffsetTFlags.py_type(fileId), 0)
def GranuleMetadataAddProductVersion(builder, productVersion): builder.PrependUOffsetTRelativeSlot(25, flatbuffers.number_types.UOffsetTFlags.py_type(productVersion), 0)
def GranuleMetadataAddLatMin(builder, latMin): builder.PrependFloat32Slot(26, latMin, 0.0)
def GranuleMetadataAddLatMax(builder, latMax): builder.PrependFloat32Slot(27, latMax, 0.0)
def GranuleMetadataAddLonMin(builder, lonMin): builder.PrependFloat32Slot(28, lonMin, 0.0)
def GranuleMetadataAddLonMax(builder, lonMax): builder.PrependFloat32Slot(29, lonMax, 0.0)
def GranuleMetadataAddCreatorEmail(builder, creatorEmail): builder.PrependUOffsetTRelativeSlot(30, flatbuffers.number_types.UOffsetTFlags.py_type(creatorEmail), 0)
def GranuleMetadataAddStationId(builder, stationId): builder.PrependUOffsetTRelativeSlot(31, flatbuffers.number_types.UOffsetTFlags.py_type(stationId), 0)
def GranuleMetadataAddPlatform(builder, platform): builder.PrependUOffsetTRelativeSlot(32, flatbuffers.number_types.UOffsetTFlags.py_type(platform), 0)
def GranuleMetadataAddSensor(builder, sensor): builder.PrependUOffsetTRelativeSlot(33, flatbuffers.number_types.UOffsetTFlags.py_type(sensor), 0)
def GranuleMetadataEnd(builder): return builder.EndObject()



# vim: ts=4:sts=4:sw=4
#
# @author: <sylvain.herledan@oceandatalab.com>
# @date: 2016-10-20
#
# This file is part of SEAScope, a 3D visualisation and analysis application
# for satellite, in-situ and numerical model data.
#
# Copyright (C) 2014-2023 OceanDataLab
#
# SEAScope is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SEAScope. If not, see <https://www.gnu.org/licenses/>.

"""
This module handles the deserialization of extraction objects
"""

import flatbuffers
import SEAScope.types.renderable_id
import SEAScope.types.gcp

import struct
import logging
import flatbuffers.compat
import flatbuffers.number_types

logger = logging.getLogger(__name__)


def serialize(builder, obj):
    """
    Not implemented
    """
    pass


def deserialize(o):
    """
    Rebuild an extraction from a FlatBuffers buffer.

    Parameters
    ----------
    buf : bytearray
        The buffer which contain the extraction object serialized with
        FlatBuffers

    Returns
    -------
    dict
        The deserialized extraction object as a dictionary.
    """
    result = {}
    result['id'] = SEAScope.types.renderable_id.deserialize(o.Id())
    result['xArity'] = o.XArity()
    result['yArity'] = o.YArity()
    result['channels'] = o.Channels()
    result['uri'] = o.Uri()
    result['start'] = o.Start()
    result['stop'] = o.Stop()

    fields_count = o.FieldsLength()
    fields = []
    for i in range(0, fields_count):
        fields.append(o.Fields(i))
    result['fields'] = fields
    shape_count = o.ShapeLength()
    shape = []
    for i in range(0, shape_count):
        gcp = SEAScope.types.gcp.deserialize(o.Shape(i))
        shape.append(gcp)
    result['shape'] = shape

    gcps_count = o.GcpsLength()
    gcps = []
    for i in range(0, gcps_count):
        gcp = SEAScope.types.gcp.deserialize(o.Gcps(i))
        gcps.append(gcp)
    result['gcps'] = gcps

    fill_values_count = o.FillValuesLength()
    fill_values = []
    for i in range(0, fill_values_count):
        fill_values.append(o.FillValues(i))
    result['fill_values'] = fill_values

    values_count = o.BufferLength()
    fmt = '<' + 'f' * values_count
    vs = struct.Struct(fmt)
    b_offset = o._tab.Offset(12)
    buffer_offset = flatbuffers.number_types.UOffsetTFlags.py_type(b_offset)
    vector_offset = o._tab.Vector(buffer_offset)
    values = vs.unpack_from(flatbuffers.compat.memoryview_type(o._tab.Bytes),
                            vector_offset)
    result['buffer'] = values

    return result



# vim: set ts=4:sts=4:sw=4
#
# @author: <sylvain.herledan@oceandatalab.com>
#
# This file is part of SEAScope, a 3D visualisation and analysis application
# for satellite, in-situ and numerical model data.
#
# Copyright (C) 2014-2023 OceanDataLab
#
# SEAScope is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SEAScope. If not, see <https://www.gnu.org/licenses/>.

"""This module provides helpers methods to interact with the SEAScope
application.
"""

import sys
import time
import struct
import socket
import logging
import calendar
import datetime
import contextlib
import SEAScope.cmds.add_collection
import SEAScope.cmds.add_variable
import SEAScope.cmds.add_granule
import SEAScope.cmds.set_rendering
import SEAScope.cmds.get_rendering
import SEAScope.cmds.set_current_datetime
import SEAScope.cmds.set_altitude
import SEAScope.cmds.look_at
import SEAScope.cmds.zoom_in
import SEAScope.cmds.zoom_out
import SEAScope.cmds.reset_camera
import SEAScope.cmds.select_variable
import SEAScope.cmds.select_variable_by_label
import SEAScope.cmds.get_variable_identifier
import SEAScope.cmds.check_schema_version

logger = logging.getLogger(__name__)


class Throttler(object):
    """Timer that provides a throttling mechanism."""

    def __init__(self):
        self.__last_upload = 0
        self.__min_delay_ms = 0

    @property
    def min_delay(self):
        """int: Minimal delay between two successive calls to
        :func:`SEAScope.upload.Throttler.apply_delay`, in milliseconds"""
        return self.__min_delay_ms

    @min_delay.setter
    def min_delay(self, value):
        logger.debug('Set min delay to {}'.format(value))
        self.__min_delay_ms = max(0, value)

    def apply_delay(self):
        """Wait until the minimal delay since last call has been reached."""
        now = datetime.datetime.utcnow()
        now_ms = time.mktime(now.timetuple()) * 1e3 + now.microsecond / 1e3
        elapsed = now_ms - self.__last_upload
        if elapsed < self.__min_delay_ms:
            to_wait_ms = self.__min_delay_ms - elapsed
            logger.debug('Delay for {}ms'.format(to_wait_ms))
            time.sleep(to_wait_ms / 1e3)
            now = datetime.datetime.utcnow()
            now_ms = time.mktime(now.timetuple()) * 1e3 + now.microsecond / 1e3
        self.__last_upload = now_ms


throttler = Throttler()
"""`SEAScope.upload.Throttler`: Global throttler object that is shared by all
the methods provided in :mod:`SEAScope.upload`. Use it to set the minimal
delay between two API calls if SEAScope is not able to handle the number of
requests sent by Python in a short period of time."""


@contextlib.contextmanager
def connect(host, port, check_schema_version=False):
    """Create a socket and connect it to the SEAScope application which listens
    on the provided IP address and port.

    Parameters
    ----------
    host : str
        IP address of the network interface that the SEAScope application
        listens to.
    port : int
        Port number that the SEAScope application listens to.

    Returns
    -------
    socket.socket
        a socket connected to SEAScope

    Example
    -------
    >>> import SEAScope.upload
    >>> with SEAScope.upload.connect('192.168.1.32', 5555) as link:
    >>>     # Pass the link object to other method so that they can communicate
    >>>     # with SEAScope
    >>>     # [...]
    """
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((host, port))

        if check_schema_version is True:
            _check_schema_version(s)

        yield s
    except:  # noqa
        _, e, _ = sys.exc_info()
        logger.error(e)
        raise
    finally:
        logger.debug('Closing connection')
        s.send(b'')
        s.shutdown(socket.SHUT_RDWR)
        s.close()


def _check_schema_version(link) -> None:
    """Compare data serialization schema versions used by the SEAScope client
    and by this package. Issue a warning in case they do not match because
    extracting objects from a binary stream using the wrong data schema could
    lead to unpredictable errors and crashes.

    Parameters
    ----------
    link: socket.socket
        Stream socket connected to SEAScope
    """
    # Request rendering config from SEAScope
    serializer = SEAScope.cmds.check_schema_version.serialize
    builder = None
    builder, serialized = serializer(builder)
    builder.Finish(serialized)
    buf = builder.Output()
    link.sendall(struct.pack('>Q', len(buf)) + buf)

    # Retrieve result
    deserializer = SEAScope.cmds.check_schema_version.deserialize

    # SEAScope versions that do not implement the checkSchemaVersion command
    # will not return anything. The command should execute very fast, so if
    # there is no response within one second it is reasonable to conclude that
    # the SEAScope client does not support this function and therefore uses
    # an old schema version
    timeout = link.gettimeout()
    link.settimeout(1.0)
    try:
        buf = link.recv(8)
    except TimeoutError:
        logger.warning('The SEAScope viewer uses an old version of the data '
                       'serialization schema compared to the one built in this'
                       ' Python package.\n'
                       'Although some commands might work, using mismatching '
                       'schema versions may lead to errors or even crashes.\n'
                       'Please use a more recent SEAScope viewer or install '
                       'the SEAScope Python package released around the same '
                       'time as the viewer to get the best compatibility '
                       'between Python and SEAScope')
        return
    finally:
        # Restore original timeout
        link.settimeout(timeout)

    msg_len = struct.unpack('<Q', buf)
    buf = read_response(link, msg_len[0])
    result = deserializer(buf)
    version_match, recommended_python_version = result
    if version_match is False:
        logger.warning('The SEAScope Python package does not use the same '
                       'data serialization schema as the SEAScope viewer.\n'
                       'Although some commands might work, using mismatching '
                       'schema versions may lead to errors or even crashes.\n'
                       'Please install SEAScope Python package version '
                       f'"{recommended_python_version}" to get the best '
                       'compatibility between Python and SEAScope.')


def read_response(link, msg_size):
    """Fetch a fixed number of bytes from the socket connected to SEAScope.

    Parameters
    ----------
    link : socket.socket
        Stream socket connected to SEAScope
    msg_size : int
        Number of bytes to fetch from the socket

    Returns
    -------
    :obj:`list` of :obj:`bytes`
        Raw data sent by SEAScope"""
    chunk_size = msg_size
    logger.debug('Message size: {}'.format(msg_size))
    already_read = 0
    result = b''
    while already_read < msg_size:
        buf = link.recv(chunk_size)
        already_read += len(buf)
        result += buf
        chunk_size = msg_size - already_read
    return result


def collection(link, _collection):
    """Add a collection to SEAScope catalogue.

    Parameters
    ----------
    link : socket.socket
        Stream socket connected to SEAScope
    _collection : dict
        Dictionary representing the collection


    Example
    -------
    >>> import SEAScope.upload
    >>> import SEAScope.lib.utils
    >>> coll_id, coll_obj = SEAScope.lib.utils.create_collection('Dummy')
    >>> print(coll_obj)
    {'id': {'granuleLevel': False,
            'sourceId': 1,
            'collectionId': 10,
            'granuleId': 0,
            'variableId': 0},
     'mustBeCurrent': False,
     'xSeamless': False,
     'ySeamless': False,
     'NEWSAligned': False,
     'label': 'Dummy',
     'tags': {},
     'variables': [],
     'variablesList': {},
     'defaultVariable': 0}
    >>> # Edit coll_obj to customize the collection
    >>> # [...]
    >>> with SEAScope.upload.connect(host, port) as link:
    >>>     SEAScope.upload.collection(link, coll_obj)
    """
    global throttler
    throttler.apply_delay()

    serializer = SEAScope.cmds.add_collection.serialize
    builder = None
    builder, serialized = serializer(builder, _collection)
    builder.Finish(serialized)
    buf = builder.Output()
    link.sendall(struct.pack('>Q', len(buf)) + buf)


def variable(link, _variable):
    """Add a variable to SEAScope catalogue.

    Parameters
    ----------
    link : socket.socket
        Stream socket connected to SEAScope
    _variable : dict
        Dictionary representing the variable

    Example
    -------
    >>> import SEAScope.upload
    >>> import SEAScope.lib.utils
    >>> collid, coll_obj = SEAScope.lib.create_collection('My collection')
    >>> var_obj = SEAScope.lib.create_variable(coll_obj, 'My variable',
    >>>                                        ['fieldA', 'fieldB'])
    >>> print(var_obj)
    {'id': 0,
     'label': 'My variable',
     'units': '',
     'fields': ['fieldA', 'fieldB'],
     'defaultRenderingMethod': 'RASTER',
     'tags': {},
     'collection': {'granuleLevel': False,
                    'sourceId': 1,
                    'collectionId': 10,
                    'granuleId': 0,
                    'variableId': 0},
     'rendering': {'rendered': True,
                   'logscale': False,
                   'min': 0,
                   'max': 1,
                   'opacity': 1.0,
                   'zindex': 500,
                   'color': [50, 12, 87],
                   'colormap': 'grayscale',
                   'renderMethod': 'RASTER',
                   'particlesCount': 1000,
                   'particleTTL': 10,
                   'streamlineLength': 20,
                   'streamlineSpeed': 0.0,
                   'filterMode': 'NEAREST',
                   'target': {'granuleLevel': False,
                              'sourceId': 1,
                              'collectionId': 10,
                              'granuleId': 0,
                              'variableId': 0}}}
    >>> # You can customize the rendering configuration for the created
    >>> # variable by altering the content of var_obj['rendering']
    >>> # [...]
    >>> with SEAScope.upload.connect('127.0.0.1', 11155) as link:
    >>>     SEAScope.upload.collection(link, coll_obj)
    >>>     SEAScope.upload.variable(link, var_obj)
    """
    global throttler
    throttler.apply_delay()

    serializer = SEAScope.cmds.add_variable.serialize
    builder = None
    collection_id = _variable['collection']
    builder, serialized = serializer(builder, collection_id, _variable)
    builder.Finish(serialized)
    buf = builder.Output()
    link.sendall(struct.pack('>Q', len(buf)) + buf)

    # Retrieve result
    deserializer = SEAScope.cmds.add_variable.deserialize
    buf = link.recv(8)
    msg_len = struct.unpack('<Q', buf)
    buf = read_response(link, msg_len[0])
    result = deserializer(buf)
    if result['ok'] is False:
        logger.error(f'Failed to add variable in SEAScope: {result["msg"]}')

    if 'rendering' in _variable:
        rendering_config(link, _variable['rendering'])


def rendering_config(link, rcfg):
    """Update a rendering configuration

    Parameters
    ----------
    link : socket.socket
        Stream socket connected to SEAScope
    rcfg : dict
        Dictionary representing the rendering configuration

    Example
    -------
    >>> import SEAScope.upload
    >>> rcfg = {'rendered': True,
    >>>         'logscale': False,
    >>>         'min': 0.0,
    >>>         'max': 1.5,
    >>>         'opacity': 0.7,
    >>>         'zindex': 0.25,
    >>>         'color': [0, 0, 0],
    >>>         'colormap': 'jet',
    >>>         'renderMethod': 'RASTER',
    >>>         'filterMode': 'BILINEAR',
    >>>         'particlesCount': 0,
    >>>         'particleTTL': 0,
    >>>         'streamlineLength': 0,
    >>>         'streamlineSpeed': 0.0,
    >>>         'target': None}
    >>> with SEAScope.upload.connect('127.0.0.1', 11155) as link:
    >>>     target = SEAScope.upload.get_id_for(link, 'My collection',
    >>>                                         'My variable')
    >>>     # Update target in the rendering configuration dictionary
    >>>     rcfg['target'] = target
    >>>     SEAScope.upload.rendering_config(link, rcfg)
    """
    global throttler
    throttler.apply_delay()

    serializer = SEAScope.cmds.set_rendering.serialize
    builder = None
    builder, serialized = serializer(builder, rcfg)
    builder.Finish(serialized)
    buf = builder.Output()
    link.sendall(struct.pack('>Q', len(buf)) + buf)


def granule(link, _granule):
    """Add a granule in SEAScope.

    Parameters
    ----------
    link : socket.socket
        Stream socket connected to SEAScope
    _granule : dict
        Dictionary representing the granule

    Example
    -------
    >>> import datetime
    >>> import SEAScope.upload
    >>> import SEAScope.lib.utils
    >>> # Example with 1D data: the GCPs will be the trajectory positions
    >>> lons = [1, 2, 3, 4, 5]  # dummy longitudes for the sake of example
    >>> lats = [|, 2, 3, 4, 5]  # dummy latitudes for the sake of example
    >>> gcps = [{'lon': lons[i], 'lat': lats[i], 'i': i, 'j': 0}
    >>>         for i in range(0, len(lons))]
    >>> # Define the time coverage of the granule
    >>> start = datetime.datetime(2019, 5, 21)
    >>> stop = datetime.datetime(2019, 5, 22, 14, 18, 56)
    >>> # Create a collection
    >>> coll_id, coll_obj = SEAScope.lib.utils.create_collection('My data')
    >>> # Create the granule object
    >>> gra_id, gra_obj = SEAScope.lib.utils.create_granule(coll_id, gcps,
    >>>                                                     start, stop)
    >>> # Associate some data with the granule
    >>> values = [31, 32, 33, 34, 35]  # replace this by actual measurements
    >>> SEAScope.lib.utils.set_field(gra_obj, 'field_name', values)
    >>> print(gra_obj)
    {'id': {'granuleLevel': True,
            'sourceId': 1,
            'collectionId': 10,
            'granuleId': 1000,
            'variableId': 0},
     'metadata': {'sourceId': 1,
                  'collectionId': 10,
                  'granuleId': 1000,
                  'dataId': 'user_generated_granule_1000',
                  'dataModel': 'TIME',
                  'start': 1558396800000,
                  'stop': 1558534736000,
                  'uris': [{'uri': 'Python: user_generated_granule_1000',
                            'xArity': 5,
                            'yArity': 2,
                            'resolution': 1000000000,
                            'subsampling_factor': 0,
                            'shape': {'xArity': 5,
                                      'yArity': 1,
                                      'gcps': [{'lon': 1, 'lat': 1, 'i': 0, 'j': 0},  # noqa:E501
                                               {'lon': 2, 'lat': 2, 'i': 1, 'j': 0},  # noqa:E501
                                               {'lon': 3, 'lat': 3, 'i': 2, 'j': 0},  # noqa:E501
                                               {'lon': 4, 'lat': 4, 'i': 3, 'j': 0},  # noqa:E501
                                               {'lon': 5, 'lat': 5, 'i': 4, 'j': 0}]}}],  # noqa:E501
                  'title': 'user_generated_granule_1000',
                  'institution': '',
                  'comment': '',
                  'file_id': 'Python: user_generated_granule_1000',
                  'product_version': '0.0.0',
                  'lat_min': 1,
                  'lat_max': 5,
                  'lon_min': 1,
                  'lon_max': 5,
                  'creator_email': '',
                  'station_id': '',
                  'platform': '',
                  'sensor': ''},
     'data': {'field_name': {'info': {'channels': 1,
                                      'xArity': 5,
                                      'yArity': 2,
                                      'dataType': 'uint8',
                                      'offsets': [31],
                             'scaleFactors': [0.015748031496062992],
                             'fillValues': [255]},
                             'buffer': [254, 190, 127, 63, 0]}}}
    >>> # Send everything to SEAScope
    >>> with SEAScope.upload.connect('127.0.0.1', 11155) as link:
    >>>     # Create the collection first!
    >>>     SEAScope.upload.collection(link, coll_obj)
    >>>     # Then the granule
    >>>     SEAScope.upload.granule(link, gra_obj)

    """
    global throttler
    throttler.apply_delay()

    serializer = SEAScope.cmds.add_granule.serialize
    builder = None
    builder, serialized = serializer(builder, _granule)
    builder.Finish(serialized)
    buf = builder.Output()
    link.sendall(struct.pack('>Q', len(buf)) + buf)


def current_datetime(link, dt):
    """Set current datetime in SEAScope.

    Parameters
    ----------
    link : socket.socket
        Stream socket connected to SEAScope
    dt : datetime.datetime
        Value that will be used as current datetime
    """
    global throttler
    throttler.apply_delay()

    timestamp = calendar.timegm(dt.timetuple())
    serializer = SEAScope.cmds.set_current_datetime.serialize
    builder = None
    builder, serialized = serializer(builder, timestamp)
    builder.Finish(serialized)
    buf = builder.Output()
    link.sendall(struct.pack('>Q', len(buf)) + buf)


def rendering_config_for(link, target):
    """Get the current rendering configuration for a (collection, variable)
    couple.

    Parameters
    ----------
    link : socket.socket
        Stream socket connected to SEAScope
    target : dict
        Dictionary which contains the numerical identifiers for the target

    Returns
    -------
    dict
        Dictionary representing the rendering configuration

    Example
    -------
    >>> import SEASope.upload
    >>> with SEAScope.upload.connect('127.0.0.1', 11155) as link:
    >>>     target = SEAScope.upload.get_id_for(link, 'AVISO altimetry',
    >>>                                         'Mean Dynamic Topography')
    >>>     rcfg = SEAScope.upload.rendering_config_for(link, target)
    >>> print(rcfg)
    {'rendered': True,
     'logscale': False,
     'min': -0.30000001192092896,
     'max': 1.5,
     'opacity': 0.8999999761581421,
     'zindex': 0.22220000624656677,
     'color': [0, 0, 0],
     'colormap': b'jet',
     'renderMethod': 'RASTER',
     'filterMode': 'BILINEAR',
     'particlesCount': 0,
     'particleTTL': 0,
     'streamlineLength': 0,
     'streamlineSpeed': 0.0,
     'target': {'granuleLevel': False,
                'sourceId': 2,
                'collectionId': 11,
                'granuleId': 0,
                'variableId': 0}}
    """
    global throttler
    throttler.apply_delay()

    # Request rendering config from SEAScope
    serializer = SEAScope.cmds.get_rendering.serialize
    deserializer = SEAScope.cmds.get_rendering.deserialize
    builder = None
    builder, serialized = serializer(builder, target)
    builder.Finish(serialized)
    buf = builder.Output()
    link.sendall(struct.pack('>Q', len(buf))+buf)

    # Retrieve result
    buf = link.recv(8)
    msg_len = struct.unpack('<Q', buf)
    buf = read_response(link, msg_len[0])
    result = deserializer(buf)
    return result


def altitude(link, altitude):
    """Set the altitude of the camera in SEAScope.

    Parameters
    ----------
    link : socket.socket
        Stream socket connected to SEAScope
    altitude : float
        Distance between the ground and the camera, in meters
    """
    global throttler
    throttler.apply_delay()

    serializer = SEAScope.cmds.set_altitude.serialize
    builder = None
    builder, serialized = serializer(builder, altitude)
    builder.Finish(serialized)
    buf = builder.Output()
    link.sendall(struct.pack('>Q', len(buf))+buf)


def location(link, lon, lat):
    """Set the location displayed at the center of the screen (the camera's
    target) in SEAScope.

    Parameters
    ----------
    link : socket.socket
        Stream socket connected to SEAScope
    lon : float
        Longitude of the location
    lat : float
        Latitude of the location

    Notes
    -----
    The values passed to this method for longitudes and latitudes are wrapped
    automatically to match the domains that SEAScope can handle
    """
    global throttler
    throttler.apply_delay()

    lon = ((lon + 180.) % 360.) - 180.
    serializer = SEAScope.cmds.look_at.serialize
    builder = None
    builder, serialized = serializer(builder, lon, lat)
    builder.Finish(serialized)
    buf = builder.Output()
    link.sendall(struct.pack('>Q', len(buf))+buf)


def zoom_in(link):
    """Zoom-in in SEAScope.
    The camera gets closer to the ground: camera altitude decreases by a
    distance that varies depending on the current altitude.

    Parameters
    ----------
    link : socket.socket
        Stream socket connected to SEAScope
    """
    global throttler
    throttler.apply_delay()

    serializer = SEAScope.cmds.zoom_in.serialize
    builder = None
    builder, serialized = serializer(builder)
    builder.Finish(serialized)
    buf = builder.Output()
    link.sendall(struct.pack('>Q', len(buf))+buf)


def zoom_out(link):
    """Zoom-out in SEAScope.
    The camera goes further away from the ground: camera altitude increases by
    a distance that varies depending on the current altitude.

    Parameters
    ----------
    link : socket.socket
        Stream socket connected to SEAScope
    """
    global throttler
    throttler.apply_delay()

    serializer = SEAScope.cmds.zoom_out.serialize
    builder = None
    builder, serialized = serializer(builder)
    builder.Finish(serialized)
    buf = builder.Output()
    link.sendall(struct.pack('>Q', len(buf))+buf)


def reset_camera(link):
    """Move camera back to its default position.

    Parameters
    ----------
    link : socket.socket
        Stream socket connected to SEAScope
    """
    global throttler
    throttler.apply_delay()

    serializer = SEAScope.cmds.reset_camera.serialize
    builder = None
    builder, serialized = serializer(builder)
    builder.Finish(serialized)
    buf = builder.Output()
    link.sendall(struct.pack('>Q', len(buf))+buf)


def select_variable_by_id(link, source_id, collection_id, variable_id,
                          selected, exclusive):
    """Update the variables selection in SEAScope catalogue using numerical
    values to identify the target variable.

    Parameters
    ----------
    link : socket.socket
        Stream socket connected to SEAScope
    source_id : int
        Numerical identifier for the data source for the collection
    collection_id : int
        Numerical identifier for the collection that contains the target
        variable
    variable_id:
        Numerical identifier for the target variable
    selected : bool
        New selection state for the variable (True = selected)
    exclusive : bool
        When set to True, this flag tells SEAScope to unselect all the
        variables that were previously selected (exclusive selection)
    """
    global throttler
    throttler.apply_delay()

    _rid = {'granuleLevel': False,
            'sourceId': source_id,
            'collectionId': collection_id,
            'granuleId': 0,
            'variableId': variable_id}

    serializer = SEAScope.cmds.select_variable.serialize
    builder = None
    builder, serialized = serializer(builder, _rid, selected, exclusive)
    builder.Finish(serialized)
    buf = builder.Output()
    link.sendall(struct.pack('>Q', len(buf))+buf)


def select_variable_by_label(link, collection_label, variable_label, selected,
                             exclusive):
    """Update the variables selection in SEAScope catalogue using numerical
    values to identify the target variable.

    Parameters
    ----------
    link : socket.socket
        Stream socket connected to SEAScope
    collection_label : str
        Label of the collection that contains the target variable
    variable_label : str
        Label of the targt variable
    selected : bool
        New selection state for the variable (True = selected)
    exclusive : bool
        When set to True, this flag tells SEAScope to unselect all the
        variables that were previously selected (exclusive selection)
    """
    global throttler
    throttler.apply_delay()

    serializer = SEAScope.cmds.select_variable_by_label.serialize
    builder = None
    builder, serialized = serializer(builder, collection_label, variable_label,
                                     selected, exclusive)
    builder.Finish(serialized)
    buf = builder.Output()
    link.sendall(struct.pack('>Q', len(buf))+buf)


def get_id_for(link, collection_label, variable_label):
    """Get the internal identifier which corresponds to a (collection,
    variable) couple using their labels (i.e. the text that describes them in
    the SEAScope catalogue menu).

    Parameters
    ----------
    link : socket
        Socket connected to the SEAScope application
    collection_label : str
        Label of the collection
    variable_label : str
        Label of the variable

    Returns
    -------
    dict
        Dictionary containing the numerical identifiers associated with the
        (collection, variable) couple

    Example
    -------
    >>> import SEAScope.upload
    >>> with SEAScope.upload.connect('127.0.0.1', 11155) as link:
    >>>     result = SEAScope.upload.get_id_for(link, 'ECMWF',
    >>>                                         'mean wind field')
    >>> print(result)
    {'granule_level': False,
     'sourceId': 2,
     'collectionId': 2,
     'granuleId': 0,
     'variableId': 0}
    """
    global throttler
    throttler.apply_delay()

    # Request rendering config from SEAScope
    serializer = SEAScope.cmds.get_variable_identifier.serialize
    deserializer = SEAScope.cmds.get_variable_identifier.deserialize
    builder = None
    builder, serialized = serializer(builder, collection_label, variable_label)
    builder.Finish(serialized)
    buf = builder.Output()
    link.sendall(struct.pack('>Q', len(buf))+buf)

    # Retrieve result
    buf = link.recv(8)
    msg_len = struct.unpack('<Q', buf)
    buf = read_response(link, msg_len[0])
    result = deserializer(buf)
    return result



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: API

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class GranuleId(object):
    __slots__ = ['_tab']

    # GranuleId
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # GranuleId
    def SourceId(self): return self._tab.Get(flatbuffers.number_types.Uint32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(0))
    # GranuleId
    def GranuleId(self): return self._tab.Get(flatbuffers.number_types.Uint64Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(8))

def CreateGranuleId(builder, sourceId, granuleId):
    builder.Prep(8, 16)
    builder.PrependUint64(granuleId)
    builder.Pad(4)
    builder.PrependUint32(sourceId)
    return builder.Offset()



# vim: ts=4:sts=4:sw=4
#
# @author: <sylvain.herledan@oceandatalab.com>
# @date: 2016-09-15
#
# This file is part of SEAScope, a 3D visualisation and analysis application
# for satellite, in-situ and numerical model data.
#
# Copyright (C) 2014-2023 OceanDataLab
#
# SEAScope is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SEAScope. If not, see <https://www.gnu.org/licenses/>.

"""
This module handles the serialization of the getRendering command and the
deserialization of the result sent by SEAScope
"""

import logging
import flatbuffers
import SEAScope.API.OpCode
import SEAScope.API.GetRendering
import SEAScope.API.GetRenderingResponse
import SEAScope.API.Command
import SEAScope.types.renderable_id
import SEAScope.types.rendering_cfg

logger = logging.getLogger(__name__)


def serialize(builder, rid):
    """
    Serialize a getRendering command using FlatBuffers.

    Parameters
    ----------
    builder : flatbuffers.builder.Builder
        The FlatBuffers builder instance which serializes data. If this
        parameter is None, then a new builder will be created
    rid : dict
        The target of the rendering configuration SEAScope must look for. The
        value must satisfy the requirements of the
        :func:`SEAScope.types.renderable_id.serialize` method

    Returns
    -------
    tuple(flatbuffers.builder.Builder, int)
        A tuple which contains two elements:

        - the :obj:`flatbuffers.builder.Builder` instance which has been used
          to serialize data
        - an :obj:`int` which is the address/offset of the serialized object
          in the builder buffer
    """
    if builder is None:
        builder = flatbuffers.Builder(0)

    _, _rid = SEAScope.types.renderable_id.serialize(builder, rid)

    cls = SEAScope.API.GetRendering
    cls.GetRenderingStart(builder)
    cls.GetRenderingAddTarget(builder, _rid)
    cmd_args = cls.GetRenderingEnd(builder)

    opcode = SEAScope.API.OpCode.OpCode().getRendering
    args_type = SEAScope.API.CommandArgs.CommandArgs().GetRendering
    SEAScope.API.Command.CommandStart(builder)
    SEAScope.API.Command.CommandAddOpcode(builder, opcode)
    SEAScope.API.Command.CommandAddArgsType(builder, args_type)
    SEAScope.API.Command.CommandAddArgs(builder, cmd_args)
    cmd = SEAScope.API.Command.CommandEnd(builder)

    return builder, cmd


def deserialize(buf):
    """
    Deserialize the response that SEAScope sends after executing the
    getRendering command

    Parameters
    ----------
    buf : bytearray
        The buffer which contains the result of the getRendering command
        serialized with FlatBuffers

    Returns
    -------
    dict
        The rendering configuration used by SEAScope for the target. The dict
        is generated by the
        :func:`SEAScope.types.rendering_cfg.deserialize` method (see source
        code for more details)
    """
    cls = SEAScope.API.GetRenderingResponse.GetRenderingResponse
    res = cls.GetRootAsGetRenderingResponse(buf, 0)

    rcfg = SEAScope.types.rendering_cfg.deserialize(res.Cfg())

    return rcfg



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: API

class DataModel(object):
    LAT_LON = 0
    Y_X = 1
    NJ_NI = 2
    ROW_CELL = 3
    TIME_STATION = 4
    TIME = 5

